#include <aperture.hpp>
#include <omp.h>
#include <complex>
using namespace QOSM::Propagation;
using namespace QOSM::EM::Source;
using namespace std::complex_literals;

void sortModes(
    int n_modes, 
    double Lx, 
    double Ly, 
    nc::NdArray<double> &k_h, 
    nc::NdArray<int> &idx_modes_h
){
    double xy_ratio = Lx / Ly;

    const auto grid = nc::meshgrid(
        nc::arange(n_modes + 1), 
        nc::arange(n_modes + 1)
    );
    const auto M = grid.first.flatten();
    const auto N = grid.second.flatten();

    // TE modes
    std::vector<std::vector<double>> k_h_arr(M.size());

    for (int n = 0; n < M.size(); n++){
        double k_val = std::sqrt(M[n] * M [n]+ xy_ratio * xy_ratio * N[n] * N[n]);
        k_h_arr.at(n) = {k_val, (double)M[n], (double)N[n]};
    }

    k_h_arr.erase(k_h_arr.begin()); // remove mode (0,0) that does not exist
    std::sort(k_h_arr.begin(), k_h_arr.end(), [](const std::vector<double>& a, const std::vector<double>& b) {
        return a[0] < b[0];
    });
    int ni = 0;
    for (auto& mode : k_h_arr) {
        if (ni >= n_modes) break;
        idx_modes_h.put(ni, 0, (int)mode[1]);
        idx_modes_h.put(ni, 1, (int)mode[2]);
        k_h.put(ni, 0, mode[1] * QOSM::EM::pi / Lx);
        k_h.put(ni, 1, mode[2] * QOSM::EM::pi / Ly);
        ni++;
    }
}


Aperture::Aperture(const double &freq, const size_t &numModes, const Frame &frame) 
: freq(freq), numModes(numModes), frame(frame){
    k0 = 2 * nc::constants::pi * freq / 299792458.;
}

nc::NdArray<cplx> Aperture::computeRx(const std::shared_ptr<Source> &src, const VectorNx3 &r, const bool &onlyCentralPoint) const{
    return {0.};
}

RectangularAperture::RectangularAperture(
    const double &freq, 
    const double &Lx, 
    const double &Ly, 
    const size_t &numModes, 
    const size_t &numPtsPerAxis, 
    const Frame &frame
): Aperture(freq, numModes, frame), numAperturePoints(numPtsPerAxis*numPtsPerAxis), Lx(Lx), Ly(Ly){

    nc::NdArray<int> mn_te(numModes, 2);
    nc::NdArray<double> kt_te(numModes, 2);
    sortModes(numModes, Lx, Ly, kt_te, mn_te);
    cplx j(0., 1.);
        
    // numModes instead of 2*numModes because only TE for now
    Ek = std::vector<FieldNx3>(numModes);
    Hk = std::vector<FieldNx3>(numModes);
    Zk = std::vector<cplx>(numModes);

    xa = nc::linspace<double>(-Lx/2., Lx/2., numPtsPerAxis, true);
    ya = nc::linspace<double>(-Ly/2., Ly/2., numPtsPerAxis, true);
    auto grid = nc::meshgrid(xa, ya);
    Xa = grid.first.flatten();
    Ya = grid.second.flatten();
    numAperturePoints = Xa.size();

    for(int k=0; k<numModes; k++){
        const int m = mn_te.at(k, 0);
        const int n = mn_te.at(k, 1);
        const double kt_te_x = kt_te.at(k, 0);
        const double kt_te_y = kt_te.at(k, 1);
        const double kch2 = kt_te_x * kt_te_x + kt_te_y * kt_te_y;

        // impedance and admintance
        const cplx Zk_ = j * k0 * Z0 / std::sqrt(kch2 - k0 * k0 + 0.*j);
        const cplx Zk_sq = std::sqrt(Zk_);
        const cplx Yk_sq = 1. / Zk_sq;

        // normalisation
        const double delta_m = 1. + (double)(m == 0);
        const double delta_n = 1. + (double)(n == 0);
        const double Nh = 1. / std::abs(kch2 * 0.25 * Lx * Ly * delta_m * delta_n);
        const double Nh_sq = std::sqrt(Nh);

        FieldNx3 ek(Xa.size());
        ek.put(ek.rSlice(), 0, -n * pi / Ly * cos(pi * m * (Xa / Lx + .5)) * sin(pi * n * (Ya / Ly + .5)) * Nh_sq * Zk_sq);
        ek.put(ek.rSlice(), 1,  m * pi / Lx * cos(pi * n * (Ya / Ly + .5)) * sin(pi * m * (Xa / Lx + .5)) * Nh_sq * Zk_sq);
        
        FieldNx3 hk(Xa.size());
        hk.put(hk.rSlice(), 0, -m * pi / Lx * cos(pi * n * (Ya / Ly + .5)) * sin(pi * m * (Xa / Lx + .5)) * Nh_sq * Yk_sq);
        hk.put(hk.rSlice(), 1, -n * pi / Ly * cos(pi * m * (Xa / Lx + .5)) * sin(pi * n * (Ya / Ly + .5)) * Nh_sq * Yk_sq);
        
        Ek.at(k) = ek;
        Hk.at(k) = hk;
        Zk.at(k) = Zk_;
    }   

}

std::tuple<FieldNx3, FieldNx3, VectorNx3> RectangularAperture::fields(
    const std::vector<cplx> modes, 
    const VectorNx3 &r_src
) const{
    const size_t numApPoints = Xa.size();
    const size_t numPoints = r_src.shape().rows;
    FieldNx3 Ea(numApPoints);
    FieldNx3 Ha(numApPoints);
    for (int k=0; k<numModes; k++){
        Ea += Ek.at(k) * modes.at(k);
        Ha += Hk.at(k) * modes.at(k);
    }

    // field computed in local frame
    const auto r_loc = frame.toLocal(r_src);
    FieldNx3 eField(numPoints);
    FieldNx3 hField(numPoints);
    VectorNx3 poyinting(numPoints);
    
    double dx = nc::abs(xa[1] - xa[0]);
    double dy = nc::abs(ya[1] - ya[0]);
    double dS = dx * dy;
    
    // consider free-space impedence so far
    double Z = 2. * nc::sqrt(nc::constants::pi * 1e-7 / 8.854187817e-12);
    double inv_Z = 1. / Z;

    const auto sl = nc::Slice(0, numApPoints);

    // Get the number of available CPU cores
    int num_cores = std::thread::hardware_concurrency();

    // Calculate the number of threads to use (85% of available cores)
    int num_hardware_threads  = static_cast<int>(num_cores * 0.85);

    // Set the number of threads for OpenMP
    omp_set_num_threads(num_hardware_threads );

    const cplx cte1 = - 1i * Z * k0 * k0 / (4 * nc::constants::pi) * dS;
    const cplx cte2 = - 1i * k0 * k0 / (4 * nc::constants::pi * Z) * dS;
    
    #pragma omp parallel for
    for(int i=0; i<numPoints; i++){
        VectorNx3 rHat(numApPoints);
       
        rHat.put(sl, 0, r_loc.at(i, 0) - Xa);
        rHat.put(sl, 1, r_loc.at(i, 1) - Ya);
        rHat.put(sl, 2, r_loc.at(i, 2));

        // normalise position vector and retrieve the norms
        const auto r = rHat.norm().reshape(-1, 1);
        rHat /= r;

        // Electric surface currents = z cross Ha
        // 0 * Ha_z - 1 * Ha_y
        // 1 * Ha_x - 0 * Ha_z
        // 0 * Ha_y - 0 * Ha_x
        auto J = nc::zeros<cplx>(numApPoints, 3);
        J.put(sl, 0, -Ha(sl, 1));
        J.put(sl, 1,  Ha(sl, 0));

        // Magnetic surface currents = - z cross Ea
        // - 0 * Ea_z + 1 * Ea_y
        // - 1 * Ea_x + 0 * Ea_z
        // - 0 * Ea_y + 0 * Ea_x
        auto M = nc::zeros<cplx>(numApPoints, 3);
        M.put(sl, 0,  Ea(sl, 1));
        M.put(sl, 1, -Ea(sl, 0));

        auto mCrossR = nc::zeros<cplx>(numApPoints, 3);
        mCrossR.put(sl, 0, M(sl, 1) * rHat(sl, 2) - M(sl, 2) * rHat(sl, 1));
        mCrossR.put(sl, 1, M(sl, 2) * rHat(sl, 0) - M(sl, 0) * rHat(sl, 2));
        mCrossR.put(sl, 2, M(sl, 0) * rHat(sl, 1) - M(sl, 1) * rHat(sl, 0));
        
        auto jCrossR = nc::zeros<cplx>(numApPoints, 3);
        jCrossR.put(sl, 0, J(sl, 1) * rHat(sl, 2) - J(sl, 2) * rHat(sl, 1));
        jCrossR.put(sl, 1, J(sl, 2) * rHat(sl, 0) - J(sl, 0) * rHat(sl, 2));
        jCrossR.put(sl, 2, J(sl, 0) * rHat(sl, 1) - J(sl, 1) * rHat(sl, 0));
        
        nc::NdArray<cplx> rDotJ = nc::sum(rHat * nc::conj(J), nc::Axis::COL).reshape(-1, 1);
        nc::NdArray<cplx> rDotM = nc::sum(rHat * nc::conj(M), nc::Axis::COL).reshape(-1, 1);
        
        // Green Function
        const auto G = nc::exp(-1.0i * k0 * r) / (k0 * r);
        const auto inv_kr = 1. / (k0 * r); 
        const auto inv_kr2 = inv_kr * inv_kr;

        const auto A = 1. - 1.0i * inv_kr - inv_kr2;
        const auto B = - 1. + 3.0i * inv_kr - 3. * inv_kr2;
        const auto C = 1. - 1.0i * inv_kr;
        
        const auto dE = (A * J + B * rDotJ * rHat + inv_Z * C * mCrossR) * G;
        const auto dH = (A * M + B * rDotM * rHat - Z * C * jCrossR) * G;
        
        // dE and dH integration along x and y
        const auto mE = cte1 * nc::sum(dE, nc::Axis::ROW);
        const auto mH = cte2 * nc::sum(dH, nc::Axis::ROW);

        // Compute Poynting vector
        auto mHc = nc::conj(mH);
        eField.put(i, 0, mE[0]);
        eField.put(i, 1, mE[1]);
        eField.put(i, 2, mE[2]);
        hField.put(i, 0, mH[0]);
        hField.put(i, 1, mH[1]);
        hField.put(i, 2, mH[2]);
        poyinting.put(i, 0, nc::real(mE[1] * mHc[2] - mE[2] * mHc[1]));
        poyinting.put(i, 1, nc::real(mE[2] * mHc[0] - mE[0] * mHc[2]));
        poyinting.put(i, 2, nc::real(mE[0] * mHc[1] - mE[1] * mHc[0]));
    }
    // fields have been compute in local frame, write them in source field
    return std::make_tuple(
        frame.toRef(eField), 
        frame.toRef(hField), 
        frame.rot2Ref(poyinting)
    );
}

FieldNx3 RectangularAperture::eField(
    const std::vector<cplx> modes, 
    const VectorNx3 &r_src
) const{
    auto res = fields(modes, r_src);
    return std::get<0>(res);
}

FieldNx3 RectangularAperture::hField(
    const std::vector<cplx> modes, 
    const VectorNx3 &r_src
) const{
    auto res = fields(modes, r_src);
    return std::get<1>(res);
}


nc::NdArray<cplx> RectangularAperture::computeRx(const FieldNx3 &E_ap_src) const{    
    // writte those fields in probe local frame
    auto E_loc = frame.toLocal(E_ap_src);

    double dx = nc::abs(xa[1] - xa[0]);
    double dy = nc::abs(ya[1] - ya[0]);
    double dS = dx * dy;

    const size_t numApPoints = Xa.size();
    const size_t numPoints = E_ap_src.numRows()/numApPoints;
    nc::NdArray<cplx> ck(numPoints, numModes);
    for (int i=0; i< numPoints; i++){
        const auto E = E_loc(nc::Slice(i*numApPoints, (i+1)*numApPoints), nc::Slice(0, 3));
        // perform modal expansion
        for (int k=0; k<numModes; k++){
            auto Ekc = nc::conj(Ek.at(k));
            const cplx Yk = 1. / Zk.at(k);
            const auto crossDotZ = Yk * nc::sum(E_loc * Ekc, nc::Axis::COL);
            ck.put(i, k, nc::sum(crossDotZ)[0] * dS);
        }
    }

    return ck;
}

nc::NdArray<cplx> RectangularAperture::computeRx(
    const std::shared_ptr<Source> &src, 
    const VectorNx3 &r_src, 
    const bool &onlyCentralPoint
) const{
    const size_t numPoints = r_src.numRows();

    nc::NdArray<cplx> ck(numPoints, numModes);

    VectorNx3 r_pts_src;
    if (onlyCentralPoint){
        r_pts_src = r_src;
    } else {
        // Write Xa and Ya in src frame in case of different orientation -> Xa, Ya, Za
        VectorNx3 r_aperture_local(numAperturePoints);  
        nc::Slice a(numAperturePoints);
        r_aperture_local.put(a, 0, Xa);
        r_aperture_local.put(a, 1, Ya);
        r_aperture_local.put(a, 2, 0);

        // write aperture point in SRC reference frame
        VectorNx3 r_aperture_src = frame.toRef(r_aperture_local);

        // calculate all point (all aperture points at each point of interest)
        r_pts_src = VectorNx3(numAperturePoints * numPoints);
        int idx = 0;
        for (int i=0; i< numPoints; i++){
            nc::Slice s(i*numAperturePoints, (i+1)*numAperturePoints);
            r_pts_src.put(s, 0, r_aperture_src(a, 0) + r_src(i, 0));
            r_pts_src.put(s, 1, r_aperture_src(a, 1) + r_src(i, 1));
            r_pts_src.put(s, 2, r_aperture_src(a, 2) + r_src(i, 2));
        }
    }

    // compute fields w.r.t. the reference frame (feeding source)
    // compute fields for all points
    const FieldNx3 &E_src = src->eField(r_pts_src);
    
    // writte those fields in probe local frame
    auto E_loc = frame.toLocal(E_src);

    double dx = nc::abs(xa[1] - xa[0]);
    double dy = nc::abs(ya[1] - ya[0]);
    double dS = dx * dy;

    #pragma omp parallel for
    for (int i=0; i< numPoints; i++){
        FieldNx3 E;
        if(onlyCentralPoint){
            E = E_loc(i, nc::Slice(0, 3)) * nc::ones<cplx>(numAperturePoints, 3);
        } else {
            E = E_loc(nc::Slice(i*numAperturePoints, (i+1)*numAperturePoints), nc::Slice(0, 3));
        }

        // perform modal expansion
        for (int k=0; k<numModes; k++){
            auto Hkc = nc::conj(Hk.at(k));       
            const auto crossDotZ = nc::sum(E.at(E.rSlice(), 0) * Hkc.at(Hkc.rSlice(), 1) - E.at(E.rSlice(), 1) * Hkc.at(Hkc.rSlice(), 0), nc::Axis::COL);
            ck.put(i, k, nc::sum(crossDotZ)[0] * dS);
        }
    }

    return ck;
}

FieldNx3 RectangularAperture::eFieldRx(
    const std::shared_ptr<Source> &src, 
    const VectorNx3 &r_src, 
    const bool &onlyCentralPoint
) const{
    const size_t numPoints = r_src.numRows();

    VectorNx3 r_pts_src;
    if (onlyCentralPoint){
        r_pts_src = r_src;
    } else {
        // Write Xa and Ya in src frame in case of different orientation -> Xa, Ya, Za
        VectorNx3 pts_aperture_local(numAperturePoints);  
        nc::Slice a(numAperturePoints);
        pts_aperture_local.put(a, 0, Xa);
        pts_aperture_local.put(a, 1, Ya);
        pts_aperture_local.put(a, 2, 0);

        // write aperture point in SRC reference frame
        VectorNx3 pts_aperture_src = frame.toRef(pts_aperture_local);

        // calculate all point (all aperture points at each point of interest)
        r_pts_src = VectorNx3(numAperturePoints * numPoints);
        int idx = 0;
        for (int i=0; i< numPoints; i++){
            nc::Slice s(i*numAperturePoints, (i+1)*numAperturePoints);
            r_pts_src.put(s, 0, pts_aperture_src(a, 0) + r_src(i, 0));
            r_pts_src.put(s, 1, pts_aperture_src(a, 1) + r_src(i, 1));
            r_pts_src.put(s, 2, pts_aperture_src(a, 2) + r_src(i, 2));
        }
    }

    // compute fields w.r.t. the reference frame (feeding source)
    // compute fields for all points
    const auto E_src = src->eField(r_pts_src);

    return E_src;
}


std::vector<bool> RectangularAperture::propagativeModes() const{
    nc::NdArray<int> mn_te(numModes, 2);
    nc::NdArray<double> kt_te(numModes, 2);
    sortModes(numModes, Lx, Ly, kt_te, mn_te);
    cplx j(0., 1.);
    std::vector<bool> res(numModes);
    const double gamma = 0;
    for(int k=0; k<numModes; k++){
        const int m = mn_te.at(k, 0);
        const int n = mn_te.at(k, 1);
        const double kt_te_x = kt_te.at(k, 0);
        const double kt_te_y = kt_te.at(k, 1);
        const double kch2 = kt_te_x * kt_te_x + kt_te_y * kt_te_y;
        res.at(k) = (kch2 - k0*k0) < 0;
    }
    return res;
}

// CIRCULAR APERTURES

CircularAperture::CircularAperture(
    const double &freq, 
    const double &radius, 
    const size_t &numModes, // only TE11 implemented so far
    const std::array<size_t,2> &numPtsPerAxis, 
    const Frame &frame
): Aperture(freq, numModes, frame), numAperturePoints(numPtsPerAxis), radius(radius){

}

std::tuple<FieldNx3, FieldNx3, VectorNx3> CircularAperture::fields(
    const std::vector<cplx> modes, 
    const VectorNx3 &r_src
) const{
    const size_t numApPoints = Xa.size();
    const size_t numPoints = r_src.shape().rows;
    FieldNx3 Ea(numApPoints);
    FieldNx3 Ha(numApPoints);
    for (int k=0; k<numModes; k++){
        Ea += Ek.at(k) * modes.at(k);
        Ha += Hk.at(k) * modes.at(k);
    }

    // field computed in local frame
    const auto r_loc = frame.toLocal(r_src);
    FieldNx3 eField(numPoints);
    FieldNx3 hField(numPoints);
    VectorNx3 poyinting(numPoints);
    
    double dx = nc::abs(xa[1] - xa[0]);
    double dy = nc::abs(ya[1] - ya[0]);
    double dS = dx * dy;
    
    // consider free-space impedence so far
    double Z = 2. * nc::sqrt(nc::constants::pi * 1e-7 / 8.854187817e-12);
    double inv_Z = 1. / Z;

    const auto sl = nc::Slice(0, numApPoints);

    // Get the number of available CPU cores
    int num_cores = std::thread::hardware_concurrency();

    // Calculate the number of threads to use (85% of available cores)
    int num_hardware_threads  = static_cast<int>(num_cores * 0.85);

    // Set the number of threads for OpenMP
    omp_set_num_threads(num_hardware_threads );

    const cplx cte1 = - 1i * k0 * k0 * Z / (4 * nc::constants::pi) * dS;
    const cplx cte2 = - 1i * k0 * k0 / (4 * nc::constants::pi * Z) * dS;
    
    #pragma omp parallel for
    for(int i=0; i<numPoints; i++){
        VectorNx3 rHat(numApPoints);
       
        rHat.put(sl, 0, Xa - r_loc.at(i, 0));
        rHat.put(sl, 1, Ya - r_loc.at(i, 1));
        rHat.put(sl, 2, - r_loc.at(i, 2));

        // normalise position vector and retrieve the norms
        const auto r = rHat.norm().reshape(-1, 1);
        rHat /= r;

        // Electric surface currents = z cross Ha
        // 0 * Ha_z - 1 * Ha_x
        // 1 * Ha_x - 0 * Ha_z
        // 0 * Ha_y - 0 * Ha_x
        auto J = nc::zeros<cplx>(numApPoints, 3);
        J.put(sl, 0, -Ha(sl, 1));
        J.put(sl, 1,  Ha(sl, 0));

        // Magnetic surface currents = - z cross Ea
        // 0 * Ea_z - 1 * Ea_x
        // 1 * Ea_x - 0 * Ea_z
        // 0 * Ea_y - 0 * Ea_x
        auto M = nc::zeros<cplx>(numApPoints, 3);
        M.put(sl, 0,  Ea(sl, 1));
        M.put(sl, 1, -Ea(sl, 0));

        auto rCrossM = nc::zeros<cplx>(numApPoints, 3);
        rCrossM.put(sl, 0, rHat(sl, 1) * M(sl, 2) - rHat(sl, 2) * M(sl, 1));
        rCrossM.put(sl, 1, rHat(sl, 2) * M(sl, 0) - rHat(sl, 0) * M(sl, 2));
        rCrossM.put(sl, 2, rHat(sl, 0) * M(sl, 1) - rHat(sl, 1) * M(sl, 0));

        auto rCrossJ = nc::zeros<cplx>(numApPoints, 3);
        rCrossJ.put(sl, 0, rHat(sl, 1) * J(sl, 2) - rHat(sl, 2) * J(sl, 1));
        rCrossJ.put(sl, 1, rHat(sl, 2) * J(sl, 0) - rHat(sl, 0) * J(sl, 2));
        rCrossJ.put(sl, 2, rHat(sl, 0) * J(sl, 1) - rHat(sl, 1) * J(sl, 0));
        
        nc::NdArray<cplx> rDotJ = nc::sum(rHat * nc::conj(J), nc::Axis::COL).reshape(-1, 1);
        nc::NdArray<cplx> rDotM = nc::sum(rHat * nc::conj(M), nc::Axis::COL).reshape(-1, 1);
        
        // Green Function
        const auto G = nc::exp(-1.0i * k0 * r) / (k0 * r);
        
        const auto inv_kr = 1. / (k0 * r); 
        const auto inv_kr2 = inv_kr * inv_kr;

        const auto A = 1. - 1.0i * inv_kr - inv_kr2;
        const auto B = - 1. + 3.0i * inv_kr - 3. * inv_kr2;
        const auto C = 1. - 1.0i * inv_kr;
        
        const auto dE = (A * J + B * rDotJ * rHat - inv_Z * C * rCrossM) * G;
        const auto dH = (A * M + B * rDotM * rHat + Z * C * rCrossJ) * G;
        
        // dE and dH integration along x and y
        const auto mE = cte1 * nc::sum(dE, nc::Axis::ROW);
        const auto mH = cte2 * nc::sum(dH, nc::Axis::ROW);

        // Compute Poynting vector
        auto mHc = nc::conj(mH);
        eField.put(i, 0, mE[0]);
        eField.put(i, 1, mE[1]);
        eField.put(i, 2, mE[2]);
        hField.put(i, 0, mH[0]);
        hField.put(i, 1, mH[1]);
        hField.put(i, 2, mH[2]);
        poyinting.put(i, 0, nc::real(mE[1] * mHc[2] - mE[2] * mHc[1]));
        poyinting.put(i, 1, nc::real(mE[2] * mHc[0] - mE[0] * mHc[2]));
        poyinting.put(i, 2, nc::real(mE[0] * mHc[1] - mE[1] * mHc[0]));
    }
    // fields have been compute in local frame, write them in source field
    return std::make_tuple(
        frame.toRef(eField), 
        frame.toRef(hField), 
        frame.rot2Ref(poyinting)
    );
}

FieldNx3 CircularAperture::eField(
    const std::vector<cplx> modes, 
    const VectorNx3 &r_src
) const{
    auto res = fields(modes, r_src);
    return std::get<0>(res);
}

FieldNx3 CircularAperture::hField(
    const std::vector<cplx> modes, 
    const VectorNx3 &r_src
) const{
    auto res = fields(modes, r_src);
    return std::get<1>(res);
}


nc::NdArray<cplx> CircularAperture::computeRx(const FieldNx3 &E_ap_src) const{    
    // writte those fields in probe local frame
    auto E_loc = frame.toLocal(E_ap_src);

    double dx = nc::abs(xa[1] - xa[0]);
    double dy = nc::abs(ya[1] - ya[0]);
    double dS = dx * dy;

    const size_t numApPoints = Xa.size();
    const size_t numPoints = E_ap_src.numRows()/numApPoints;
    nc::NdArray<cplx> ck(numPoints, numModes);
    for (int i=0; i< numPoints; i++){
        const auto E = E_loc(nc::Slice(i*numApPoints, (i+1)*numApPoints), nc::Slice(0, 3));
        // perform modal expansion
        for (int k=0; k<numModes; k++){
            auto Ekc = nc::conj(Ek.at(k));
            const cplx Yk = 1. / Zk.at(k);
            const auto crossDotZ = Yk * nc::sum(E_loc * Ekc, nc::Axis::COL);
            ck.put(i, k, nc::sum(crossDotZ)[0] * dS);
        }
    }

    return ck;
}

nc::NdArray<cplx> CircularAperture::computeRx(
    const std::shared_ptr<Source> &src, 
    const VectorNx3 &r_src, 
    const bool &onlyCentralPoint
) const{
    const size_t numPoints = r_src.numRows();
    nc::NdArray<cplx> ck(numPoints, numModes);

    return ck;
}

FieldNx3 CircularAperture::eFieldRx(
    const std::shared_ptr<Source> &src, 
    const VectorNx3 &r_src, 
    const bool &onlyCentralPoint
) const{
    const size_t numPoints = r_src.numRows();

    VectorNx3 r_pts_src;
    const auto E_src = src->eField(r_pts_src);

    return E_src;
}


std::vector<bool> CircularAperture::propagativeModes() const{
    cplx j(0., 1.);
    std::vector<bool> res(numModes);
    return res;
}