#include "bvh.hpp"

using namespace QOSM::Item;

/**
 * Create a node.
 *
 * @param bounding_box Min and max corner position vectors, representing the bounding volume that encloses all the primitives or child nodes contained within it.
 * @param left Child node on the left of the split axis.
 * @param right Child node on the right of the split axis.
 * @param triangles Triangles located in the bounding box.
 * @return The created BVH node.
 */
BVHNode::BVHNode(Vec3 bb[2], std::shared_ptr<BVHNode> left_ptr, std::shared_ptr<BVHNode> right_ptr, std::vector<triangle_ptr> tri) :
        bounding_box{bb[0], bb[1]}, left(left_ptr), right(right_ptr), triangles(tri) {}

bool BVHNode::bounding_box_intersect(const QOSM::Propagation::Ray& ray) {
    Vec3 ray_ori = ray.ori;
    Vec3 ray_dir = ray.dir;

    // Compute slab intersection distances
    Vec3 _t_near = (bounding_box[0] - ray_ori) / ray_dir;
    Vec3 _t_far = (bounding_box[1] - ray_ori) / ray_dir;

    // Sort intersection distances
    Vec3 t_near = _t_near.minimum(_t_far);
    Vec3 t_far = _t_near.maximum(_t_far);

    // Find the largest t_near and the smallest t_far
    float t_enter = t_near.max();
    float t_exit = t_far.min();

    // Check if intersection exists
    return (t_exit >= t_enter) && (t_exit >= 0);
}

/**
 * Build the BVH recursively.
 *
 * @param triangles All triangles to be considered.
 * @param depth Current depth of the BVH in construction.
 * @return The root BVH Node, nullptr if no triangles have been found.
 */
std::shared_ptr<BVHNode> QOSM::Item::build_bvh(const std::vector<triangle_ptr>& triangles, int depth){
    if (triangles.empty())
        return nullptr;

    // Calculate the bounding box of all triangles
    Vec3 min_corner = triangles[0]->v0;
    Vec3 max_corner = triangles[0]->v0;

    for (const auto& triangle : triangles) {
        min_corner = min_corner.minimum(triangle->v0).minimum(triangle->v1).minimum(triangle->v2);
        max_corner = max_corner.maximum(triangle->v0).maximum(triangle->v1).maximum(triangle->v2);
    }

    Vec3 bounding_box[2] = {min_corner, max_corner};

    // Create BVH node
    std::shared_ptr<BVHNode> node = std::make_unique<BVHNode>(bounding_box, nullptr, nullptr, triangles);

    if (triangles.size() == 1)
        return node;

    // Determine the axis with the largest extent
    int axis = (max_corner - min_corner).argmin();

    // Sort triangles based on their centroids along the determined axis
    std::vector<triangle_ptr> sorted_triangles = triangles;
    std::sort(sorted_triangles.begin(), sorted_triangles.end(), [axis](const triangle_ptr& t1, const triangle_ptr& t2) {
        return (t1->v0[axis] + t1->v1[axis] + t1->v2[axis]) / 3 < (t2->v0[axis] + t2->v1[axis] + t2->v2[axis]) / 3;
    });

    // Partition triangles into two halves along the sorted axis
    int half = sorted_triangles.size() / 2;
    const std::vector<triangle_ptr> left_triangles(sorted_triangles.begin(), sorted_triangles.begin() + half);
    const std::vector<triangle_ptr> right_triangles(sorted_triangles.begin() + half, sorted_triangles.end());

    // Recursively build BVH nodes for left and right halves
    node->left = build_bvh(left_triangles, depth + 1);
    node->right = build_bvh(right_triangles, depth + 1);

    // Return the root BVH node
    return node;
}

/**
 * Traverse the BVH to find an intersection with the ray.
 *
 * @param ray The ray as a tuple (origin, direction).
 * @param bvh_node The BVH Node potentially traversed by the ray. If it is a leaf, a precise intersection test is done with its triangles. If it is an internal node, a recursive call with its children will be carried out.
 * @return An Intersection (Triangle, distance, u, v, valid) 
 */
Intersection QOSM::Item::traverse_bvh_policy0(const QOSM::Propagation::Ray& ray, const std::shared_ptr<BVHNode> &bvh_node){
    Intersection intersection = {nullptr, 1e30, 0., 0., false};
    if (!bvh_node)
        return intersection;
    if (!bvh_node->bounding_box_intersect(ray))
        return intersection;
    if (!bvh_node->left && !bvh_node->right) {
        for (auto& triangle : bvh_node->triangles) {
            //if (triangle->idTriangle == ray.id_triangle) continue;
            double u, v, t;
            if (triangle->intersect(ray, t, u, v)) {
                intersection = {triangle, t, u, v, true};
                return intersection;
            }
        }
    } else {
        Intersection intersection1 = traverse_bvh_policy0(ray, bvh_node->left);
        Intersection intersection2 = traverse_bvh_policy0(ray, bvh_node->right);
        if (intersection1.valid && (!intersection2.valid || intersection1.t < intersection2.t)) {
            return intersection1;
        } else if (intersection2.valid) {
            return intersection2;
        }
    }
    return intersection;
}

/**
 * Traverse the BVH to find potential intersections with the ray.
 *
 * @param ray The ray as a tuple (origin, direction).
 * @param bvh_node The BVH Node potentially traversed by the ray. If it is a leaf, a precise intersection test is done with its triangles. If it is an internal node, a recursive call with its children will be carried out.
 * @return A vector containing each intersection found (Triangle, distance, u, v, valid).
 */
std::vector<Intersection> QOSM::Item::traverse_bvh_policy1(const QOSM::Propagation::Ray& ray, const std::shared_ptr<BVHNode> &bvh_node) {
    std::vector<Intersection> intersections;
    if (!bvh_node)
        return intersections;
    if (!bvh_node->bounding_box_intersect(ray))
        return intersections;
    if (!bvh_node->left && !bvh_node->right) {
        for (auto& triangle : bvh_node->triangles) {
            //if (triangle->idTriangle == ray.id_triangle) continue;
            double u, v, t;
            if (triangle->intersect(ray, t, u, v)) {
                intersections.push_back({triangle, t, u, v, true});
            }
        }
    } else {
        auto intersections1 = traverse_bvh_policy1(ray, bvh_node->left);
        auto intersections2 = traverse_bvh_policy1(ray, bvh_node->right);
        intersections.insert(intersections.end(), intersections1.begin(), intersections1.end());
        intersections.insert(intersections.end(), intersections2.begin(), intersections2.end());
    }
    return intersections;
}

/**
 * Find a potential intersection between the ray and the BVH.
 *
 * @param ray The ray as a tuple (origin, direction).
 * @param bvhRoot The BVH root node, describing the entire geometry of the scene to be considered.
 * @param policy If 0, return the first intersection found, assuming it is necessarily the closest one. If 1, return all intersections found and choose the closest one.
 * @return An Intersection (Triangle, distance, u, v, valid) 
 */
Intersection QOSM::Item::ray_intersect_bvh(const QOSM::Propagation::Ray& ray, const std::shared_ptr<BVHNode> &bvhRoot, int policy) {
    if (policy == 0){
        auto intersection = traverse_bvh_policy0(ray, bvhRoot);
        return intersection;
    } else {
        // Traverse the BVH to find potential intersections
        std::vector<Intersection> intersections = traverse_bvh_policy1(ray, bvhRoot);

        // If no intersections found, return an invalid intersection
        if (intersections.empty()) {
            return {nullptr, 1e30, 0.0, 0.0, false};
        }

        // Find the closest intersection
        auto closest_intersection = std::min_element(intersections.begin(), intersections.end(),
            [](const Intersection& lhs, const Intersection& rhs) {
                return lhs.t < rhs.t;
            });

        // Return the closest intersection
        return *closest_intersection;
    }
}