#include <beam.hpp>
#include <field.hpp>

using namespace QOSM::Propagation;
using namespace QOSM::Pose;
using namespace QOSM::EM;

Beam::Beam(){
    frame;
    phi0 = a0 = 0.;
    k0 = n = kappa= initialPower = 0;
    Z_, q1_, q2_, theta_= 0.;
    z0 = {0., 0.};
    alpha_ = {0., 1.};
    zRange = {-1e30, 1e30};
    bouncesLeft = {0, 0};
    beamType = BeamType::INITIAL;
    trck_id = -1;
    fresnel_ = {1., 1.};
    id_triangle = -1;
}

Beam::Beam(
    const double &freq, 
    const array2d &ior, 
    const array2c &polarisation, 
    const array2d &w0, 
    const array2d &z0, 
    const double &a0, 
    const double &phi0, 
    const double &initPw, 
    const array2d &zRange, 
    const array2i &bounces, 
    const Frame &frame, 
    const  int &tracking_id)
: frame(frame), n(ior[0]), kappa(ior[1]), a0(a0), phi0(phi0), z0(z0), theta_(0.), alpha_(polarisation), zRange(zRange), bouncesLeft(bounces), trck_id(tracking_id){
    k0 = 2*pi * freq / c0;
    cplx j(0., 1.);
    Z_ = Z0 / (n - kappa * j);

    // compute q1 and q2
    cplx k = k0 * (n - kappa * cplx(0., 1.));
    q1_ = - (k * w0[0] * w0[0]) / (2.*j) - z0[0];
    q2_ = - (k * w0[1] * w0[1]) / (2.*j) - z0[1];
    theta_ = 0.;

    if (initPw == 0.)
        initialPower = std::pow(std::abs(polarisation[0] * a0), 2.) + std::pow(std::abs(polarisation[1] * a0), 2.);
    beamType = BeamType::INITIAL;
    fresnel_ = {1., 1.};
    id_triangle = -1;
}

Beam::Beam(
    const double &freq, 
    const array2d &ior, 
    const array2c &polarisation, 
    const array2c &q, 
    const cplx &theta, 
    const double &a0, 
    const double &phi0, 
    const double &initPw, 
    const array2d &zRange, 
    const array2i &bounces, 
    const Frame &frame, 
    const  int &tracking_id)
: frame(frame), n(ior[0]), kappa(ior[1]), a0(a0), phi0(phi0), q1_(q[0]), q2_(q[2]), theta_(theta), alpha_(polarisation), zRange(zRange), bouncesLeft(bounces), trck_id(tracking_id){
    k0 = 2*nc::constants::pi * freq / 299792458.;
    Z_ = Z0 / (n - kappa * cplx(0., 1.));

    if (initPw == 0.)
        initialPower = std::pow(std::abs(polarisation[0] * a0), 2.) + std::pow(std::abs(polarisation[1] * a0), 2.);
    beamType = BeamType::INITIAL;
    fresnel_ = {1., 1.};
    id_triangle = -1;
}


Ray Beam::getRay(){
    Ray ray;
    ray.ori = frame.ori;
    ray.dir = frame.z;
    ray.id_triangle = id_triangle;
    return ray;
}
        
std::pair<std::array<cplx, 3>, std::array<cplx, 3>> Beam::computeFields(const Vec3 &r_pt_src, const bool noZ) const{
    Vec3 r_pt_loc = frame.toLocal(r_pt_src);
    const double &x = r_pt_loc[0];
    const double &y = r_pt_loc[1];
    const double &z = r_pt_loc[2];
    if (r_pt_loc[2] < zRange[0] || r_pt_loc[2] > zRange[1]){
        return {{0., 0., 0.}, {0., 0., 0.}};
    }

    std::array<cplx, 4> Q_;
    auto res = gaussianScalarFn(Q_, *this, r_pt_loc);
    cplx u_ = res[0];

    auto Qd_ = .5*(Q_[1] + Q_[2]);
    auto dux_ = Q_[0] * x + Qd_ * y;
    auto duy_ = Q_[3] * y + Qd_ * x;

    cplx Ez = 0;
    cplx Hz = 0;
    if (!noZ) {
        Ez = - (alpha_[0] * dux_ + alpha_[1] * duy_) * u_;
    }
    Hz = (alpha_[1] * dux_ - alpha_[0] * duy_) * u_ / Z_;

    auto e = frame.rot2Ref({
        alpha_[0] * u_,
        alpha_[1] * u_,
        Ez 
    });

    auto h = frame.rot2Ref({
        - alpha_[1] * u_ / Z_,
        alpha_[0] * u_ / Z_,
        Hz
    });

    return {e, h};
};

void  QOSM::Propagation::initBeam(Beam &beam, const std::array<cplx, 4> &Q_){
    // Compute eigenvalues of Q
    cplx t = Q_[0] + Q_[3];
    cplx d = Q_[0] * Q_[3] - Q_[1] * Q_[2];
    array2c ev = {.5 * t + std::sqrt(.25 * t * t - d), .5 * t - std::sqrt(.25 * t * t - d)};
    array2d g = {-ev[0].real() / ev[0].imag(), -ev[1].real() / ev[1].imag()};
    if (std::abs(Q_[1] + Q_[2]) >= 1e-10){
        // GAGB
        beam.theta_ = 0.5 * std::atan((Q_[1] + Q_[2]) / (Q_[1] - Q_[3]));
        beam.q1_ = 1. / ev[0];
        beam.q2_ = 1. / ev[1];
    } else {
        // SAGB
        beam.theta_ = 0.;
        beam.q1_ = 1. / Q_[0];
        beam.q2_ = 1. / Q_[3];
    }

    // [m] Beam's origins wrt the pt where QI is defined
    array2d temp = {(1./beam.q1_).real() * (1. + g[0] * g[0]),  (1./beam.q2_).real() * (1. + g[1] * g[1])};
    if (temp[0] * temp[1] != 0.){
        beam.z0 = {-g[0]*g[0] / temp[0], -g[1]*g[1] / temp[1]};
    } else {
        beam.z0 = {0., 0.};
    }
}

void  QOSM::Propagation::initBeam(Beam &beam, const std::array<cplx, 4> &Q_, const cplx &ampl, const cplx &phase){
    // Compute eigenvalues of Q
    const cplx tr = Q_[0] + Q_[3];
    const cplx det = Q_[0] * Q_[3] - Q_[1] * Q_[2];
    const cplx discriminant = tr * tr - 4. * det;
    
    const cplx lambda1 = (tr + std::sqrt(discriminant)) * .5;
    const cplx lambda2 = (tr - std::sqrt(discriminant)) * .5;

    array2c ev = {lambda1, lambda2};
    if (Q_[1] == 0. && Q_[2] == 0. && Q_[0].real() > Q_[3].real() && Q_[0].imag() > Q_[3].imag())
        ev = {lambda2, lambda1};
        
    const array2d g = {-ev[0].real() / ev[0].imag(), -ev[1].real() / ev[1].imag()};

    cplx q1_ = 1. / Q_[0];
    cplx q2_ = 1. / Q_[3];
    cplx Q0_, Q1_, Q2_, Q3_;

    if (std::abs(Q_[1] + Q_[2]) >= 1e-10){
        // GAGB
        beam.theta_ = 0.5 * std::atan((Q_[1] + Q_[2]) / (Q_[0] - Q_[3]));
        q1_ = 1. / ev[0];
        q2_ = 1. / ev[1];
        auto ctheta = std::cos(beam.theta_);
        auto stheta = std::sin(beam.theta_);
        auto s2theta = std::sin(2.*beam.theta_);
        
        Q0_ = ctheta*ctheta / q1_ + stheta*stheta / q2_;
        Q3_ = stheta*stheta / q1_ + ctheta*ctheta / q2_;
        Q1_ = Q2_ = .5 * s2theta * (1. / q1_ - 1. / q2_);
    } else {
        // SAGB
        beam.theta_ = 0.;
        Q0_ = 1. / q1_;
        Q3_ = 1. / q2_;
        Q1_ = Q2_ = 0;
    }
    beam.q1_ = q1_;
    beam.q2_ = q2_;

    // [m] Beam's origins wrt the pt where QI is defined
    const array2d temp = {(1./beam.q1_).real() * (1. + g[0] * g[0]),  (1./beam.q2_).real() * (1. + g[1] * g[1])};
    if (temp[0] * temp[1] != 0.){
        beam.z0 = {-g[0]*g[0] / temp[0], -g[1]*g[1] / temp[1]};
    } else {
        beam.z0 = {0., 0.};
    }
    const double k = beam.k0 * beam.n;
    const array2d wz = {std::sqrt(-2. / (k * std::imag(1./q1_))), 
                        std::sqrt(-2. / (k * std::imag(1./q2_)))};
    const array2d w0 = {std::sqrt(wz[0] / (1 + g[0]*g[0])), std::sqrt(wz[1] / (1 + g[1]*g[1]))};
    beam.zr = {w0[0]*w0[0] * k / 2, w0[1]*w0[1] * k / 2};

    // matching term, power-normalised
    const double g1 = Q0_.imag();
    const double g2 = Q3_.imag();
    const double g0 = (Q1_ + Q2_).imag();
    
    const double G = std::abs(4*g1 * g2 - g0 * g0);
    const double a0 = std::sqrt(.5 * beam.k0 * Z0 * std::sqrt(G) / pi);

    beam.a0 = ampl / a0;
    
    const auto eta = .5 * (
       std::atan2(q1_.real(), q1_.imag())
        + std::atan2(q2_.real(), q2_.imag())
    );

    // Gouy phase is from waist position
    /*const auto eta = .5 * (
        std::atan((- beam.z0[0]) / beam.zr[0]) + std::atan((- beam.z0[1]) / beam.zr[1])
    );*/

    beam.phi0 = phase - eta;
}

std::array<cplx, 3> QOSM::Propagation::gaussianScalarFn(std::array<cplx, 4> &Q_, const Beam &beam, const Vec3 &r_pt_loc){
    const auto x = r_pt_loc[0];
    const auto y = r_pt_loc[1];
    const auto z = r_pt_loc[2];

    const auto q1_ = beam.q1_ + z;
    const auto q2_ = beam.q2_ + z;

    // Curvature matrix
    if (beam.theta_ != 0.){
        const auto ctheta = std::cos(beam.theta_);
        const auto stheta = std::sin(beam.theta_);
        const auto s2theta = std::sin(2.*beam.theta_);
        
        Q_[0] = ctheta*ctheta / q1_ + stheta*stheta / q2_;
        Q_[3] = stheta*stheta / q1_ + ctheta*ctheta / q2_;
        Q_[1] = Q_[2] = .5 * s2theta * (1. / q1_ - 1. / q2_);
    } else {
        Q_[0] = 1. / q1_; 
        Q_[3] = 1. / q2_; 
        Q_[1] = Q_[2] = 0.;
    }

    // Amplitude normalisation
    const auto g1 = Q_[0].imag();
    const auto g2 = Q_[3].imag();
    const auto g0 = (Q_[1] + Q_[2]).imag();

    auto k = beam.k0 * beam.n;

    // G = 4 * g1 * g2 - g0 ** 2
    // sqrt( .5 * beam.k * sqrt(G) / pi)
    const auto G = std::abs(4*g1 * g2 - g0 * g0);
    const double u0_ = std::sqrt(.5 * beam.k0 * Z0 * std::sqrt(G) / nc::constants::pi);

    // Gouy phase
    // Note: Gouy phase is from waist position
    const auto eta = .5 * (
        std::atan2(q1_.real(), q1_.imag())
        + std::atan2(q2_.real(), q2_.imag())
    );
    /*const auto eta = .5 * (
        std::atan((z - beam.z0[0]) / beam.zr[0]) + std::atan((z - beam.z0[1]) / beam.zr[1])
    );*/

    //const auto absorption = - beam.kappa * beam.k0 * z;
    const auto j = cplx(0. , 1.);
    const auto gamma = beam.k0 * (beam.n - j*beam.kappa);

    // Accumulated phase (propagation)
    const auto phi_ac_ = - gamma * z;

    // complex phase term
    const auto phase = eta + phi_ac_ + beam.phi0;

    // [x y]*Qz*[x;y] -> quadratic part
    const auto rQr = Q_[0] * x * x + Q_[3] * y * y + (Q_[1] + Q_[2]) * x * y;

    // Scalar Gaussian Function
    const auto K = - .5*j * gamma * rQr;
    const auto part1 = std::exp(K + j * phase);
    
    const auto u_ = u0_ * beam.a0 * part1;

    return {std::abs(u_) > 1e-7 ? u_ : 0., u0_ * beam.a0, phase};
}
