#include <frame.hpp>
using namespace QOSM::Pose;

Frame::Frame(Vec3 x, Vec3 y, Vec3 z, Vec3 ori): ori(ori), x(x), y(y), z(z){
    R[0] = x[0]; R[1] = x[1]; R[2] = x[2];
    R[3] = y[0]; R[4] = y[1]; R[5] = y[2];
    R[6] = z[0]; R[7] = z[1]; R[8] = z[2];
};

Frame::Frame(std::initializer_list<double> values){
    /*
    * Size: 6 => pose as (x, y, z, rx, ry, rz)
    * Size: 7 => pose as (x, y, z, qw, qx, qy, qz)
    */
    auto it = values.begin();
    ori = Vec3(*it++, *it++, *it++);
    double q0, q1, q2, q3 = 0.;

    if (values.size() == 6){
        // pose = position + rv
        Vec3 axis(*it++, *it++, *it++);
        const double angle = axis.normalise();
        const double sAngle = std::sin(angle);
        q0 = std::cos(angle);
        q1 = axis[0] * sAngle;
        q2 = axis[1] * sAngle;
        q3 = axis[2] * sAngle;
    } else if (values.size() == 7){
        // pose = position + quaternion
        q0 = *it++;
        q1 = *it++;
        q2 = *it++;
        q3 = *it++;
    }
    x = Vec3(
        2 * (q0 * q0 + q1 * q1) - 1,
        2 * (q1 * q2 + q0 * q3),
        2 * (q1 * q3 - q0 * q2)
    );
    y = Vec3(
        2 * (q1 * q2 - q0 * q3),
        2 * (q0 * q0 + q2 * q2) - 1,
        2 * (q2 * q3 + q0 * q1)
    );
    z = Vec3(
        2 * (q1 * q3 + q0 * q2),
        2 * (q2 * q3 - q0 * q1),
        2 * (q0 * q0 + q3 * q3) - 1
    );
    R[0] = x[0]; R[1] = x[1]; R[2] = x[2];
    R[3] = y[0]; R[4] = y[1]; R[5] = y[2];
    R[6] = z[0]; R[7] = z[1]; R[8] = z[2];
}

Frame::Frame(const Vec3 &ori, const Quaternion &q) : ori(ori){
    const double& q0 = q[0];
    const double& q1 = q[1];
    const double& q2 = q[2];
    const double& q3 = q[3];
    x = Vec3(
        2 * (q0 * q0 + q1 * q1) - 1,
        2 * (q1 * q2 + q0 * q3),
        2 * (q1 * q3 - q0 * q2)
    );
    y = Vec3(
        2 * (q1 * q2 - q0 * q3),
        2 * (q0 * q0 + q2 * q2) - 1,
        2 * (q2 * q3 + q0 * q1)
    );
    z = Vec3(
        2 * (q1 * q3 + q0 * q2),
        2 * (q2 * q3 - q0 * q1),
        2 * (q0 * q0 + q3 * q3) - 1
    );
    R[0] = x[0]; R[1] = x[1]; R[2] = x[2];
    R[3] = y[0]; R[4] = y[1]; R[5] = y[2];
    R[6] = z[0]; R[7] = z[1]; R[8] = z[2];
}

Frame::Frame(const Pose &rpose) : ori(rpose.ref_r_ref_ori){
    const double& q0 = rpose.q_ref_ori[0];
    const double& q1 = rpose.q_ref_ori[1];
    const double& q2 = rpose.q_ref_ori[2];
    const double& q3 = rpose.q_ref_ori[3];
    x = Vec3(
        2 * (q0 * q0 + q1 * q1) - 1,
        2 * (q1 * q2 + q0 * q3),
        2 * (q1 * q3 - q0 * q2)
    );
    y = Vec3(
        2 * (q1 * q2 - q0 * q3),
        2 * (q0 * q0 + q2 * q2) - 1,
        2 * (q2 * q3 + q0 * q1)
    );
    z = Vec3(
        2 * (q1 * q3 + q0 * q2),
        2 * (q2 * q3 - q0 * q1),
        2 * (q0 * q0 + q3 * q3) - 1
    );
    R[0] = x[0]; R[1] = x[1]; R[2] = x[2];
    R[3] = y[0]; R[4] = y[1]; R[5] = y[2];
    R[6] = z[0]; R[7] = z[1]; R[8] = z[2];
}

Frame::Frame(const Vec3 &ori, const Vec3 &axis, const double &angle, const bool &deg) : ori(ori){
    const double _angle = (deg ? angle * nc::constants::pi / 180. : angle) * .5;
    const double sAngle = std::sin(_angle);
    const double q0 = std::cos(_angle);
    const double q1 = axis[0] * sAngle;
    const double q2 = axis[1] * sAngle;
    const double q3 = axis[2] * sAngle;
    x = Vec3(
        2 * (q0 * q0 + q1 * q1) - 1,
        2 * (q1 * q2 + q0 * q3),
        2 * (q1 * q3 - q0 * q2)
    );
    y = Vec3(
        2 * (q1 * q2 - q0 * q3),
        2 * (q0 * q0 + q2 * q2) - 1,
        2 * (q2 * q3 + q0 * q1)
    );
    z = Vec3(
        2 * (q1 * q3 + q0 * q2),
        2 * (q2 * q3 - q0 * q1),
        2 * (q0 * q0 + q3 * q3) - 1
    );
    R[0] = x[0]; R[1] = x[1]; R[2] = x[2];
    R[3] = y[0]; R[4] = y[1]; R[5] = y[2];
    R[6] = z[0]; R[7] = z[1]; R[8] = z[2];
}

Vec3 Frame::toLocal(const Vec3 &v) const{
    auto v_ = v - ori;
    return Vec3(
        v_[0] * R[0] + v_[1] * R[1] + v_[2] * R[2], 
        v_[0] * R[3] + v_[1] * R[4] + v_[2] * R[5],
        v_[0] * R[6] + v_[1] * R[7] + v_[2] * R[8]
    );
}
    
Vec3 Frame::toRef(const Vec3 &v) const{
    return Vec3(
        v[0] * R[0] + v[1] * R[3] + v[2] * R[6], 
        v[0] * R[1] + v[1] * R[4] + v[2] * R[7],
        v[0] * R[2] + v[1] * R[5] + v[2] * R[8]
    ) + ori;
}
Vec3 Frame::rot2Local(const Vec3 &v) const{
    return Vec3(
        v[0] * R[0] + v[1] * R[1] + v[2] * R[2], 
        v[0] * R[3] + v[1] * R[4] + v[2] * R[5],
        v[0] * R[6] + v[1] * R[7] + v[2] * R[8]
    );
}
    
Vec3 Frame::rot2Ref(const Vec3 &v) const{
    return Vec3(
        v[0] * R[0] + v[1] * R[3] + v[2] * R[6], 
        v[0] * R[1] + v[1] * R[4] + v[2] * R[7],
        v[0] * R[2] + v[1] * R[5] + v[2] * R[8]
    );
}

VectorNx3 Frame::toLocal(const VectorNx3 &v) const{
    auto v_ = v - ori.toNd(v.numRows());
    VectorNx3 res(v.numRows());
    auto r = v.rSlice();
    res.put(r, 0, v_(r, 0) * R[0] + v_(r, 1) * R[1] + v_(r, 2) * R[2]);
    res.put(r, 1, v_(r, 0) * R[3] + v_(r, 1) * R[4] + v_(r, 2) * R[5]);
    res.put(r, 2, v_(r, 0) * R[6] + v_(r, 1) * R[7] + v_(r, 2) * R[8]);
    return res;
}
    
VectorNx3 Frame::toRef(const VectorNx3 &v) const{
    VectorNx3 res(v.numRows());
    auto r = v.rSlice();
    res.put(r, 0, v(r, 0) * R[0] + v(r, 1) * R[3] + v(r, 2) * R[6]);
    res.put(r, 1, v(r, 0) * R[1] + v(r, 1) * R[4] + v(r, 2) * R[7]);
    res.put(r, 2, v(r, 0) * R[2] + v(r, 1) * R[5] + v(r, 2) * R[8]);
    return res + ori.toNd(v.numRows());
}

// for complex data
std::array<std::complex<double>, 3> Frame::rot2Local(const std::array<std::complex<double>, 3> &v) const{
    return {
        v[0] * R[0] + v[1] * R[1] + v[2] * R[2],
        v[0] * R[3] + v[1] * R[4] + v[2] * R[5],
        v[0] * R[6] + v[1] * R[7] + v[2] * R[8]
    };
}
    
std::array<std::complex<double>, 3> Frame::rot2Ref(const std::array<std::complex<double>, 3> &v) const{
    return {
        v[0] * R[0] + v[1] * R[3] + v[2] * R[6],
        v[0] * R[1] + v[1] * R[4] + v[2] * R[7],
        v[0] * R[2] + v[1] * R[5] + v[2] * R[8]
    };
}
VectorNx3 Frame::rot2Local(const VectorNx3 &v) const{
    VectorNx3 res(v.numRows());
    auto r = v.rSlice();
    res.put(r, 0, v(r, 0) * R[0] + v(r, 1) * R[1] + v(r, 2) * R[2]);
    res.put(r, 1, v(r, 0) * R[3] + v(r, 1) * R[4] + v(r, 2) * R[5]);
    res.put(r, 2, v(r, 0) * R[6] + v(r, 1) * R[7] + v(r, 2) * R[8]);
    return res;
}
    
VectorNx3 Frame::rot2Ref(const VectorNx3 &v) const{
    VectorNx3 res(v.numRows());
    auto r = v.rSlice();
    res.put(r, 0, v(r, 0) * R[0] + v(r, 1) * R[3] + v(r, 2) * R[6]);
    res.put(r, 1, v(r, 0) * R[1] + v(r, 1) * R[4] + v(r, 2) * R[7]);
    res.put(r, 2, v(r, 0) * R[2] + v(r, 1) * R[5] + v(r, 2) * R[8]);
    return res;
}

QOSM::EM::FieldNx3 Frame::toLocal(const QOSM::EM::FieldNx3 &f) const{
    EM::FieldNx3 res(f.numRows());
    auto r = f.rSlice();
    res.put(r, 0, f(r, 0) * R[0] + f(r, 1) * R[1] + f(r, 2) * R[2]);
    res.put(r, 1, f(r, 0) * R[3] + f(r, 1) * R[4] + f(r, 2) * R[5]);
    res.put(r, 2, f(r, 0) * R[6] + f(r, 1) * R[7] + f(r, 2) * R[8]);
    return res;
}
    
QOSM::EM::FieldNx3 Frame::toRef(const QOSM::EM::FieldNx3 &f) const{
    EM::FieldNx3 res(f.numRows());
    auto r = f.rSlice();
    res.put(r, 0, f(r, 0) * R[0] + f(r, 1) * R[3] + f(r, 2) * R[6]);
    res.put(r, 1, f(r, 0) * R[1] + f(r, 1) * R[4] + f(r, 2) * R[7]);
    res.put(r, 2, f(r, 0) * R[2] + f(r, 1) * R[5] + f(r, 2) * R[8]);
    return res;
}