#include <pybind11/pybind11.h>
#include <pybind11/numpy.h>
#include <pybind11/stl.h>
#include <beam.hpp>
#include <propagation.hpp>
#include <triangle.hpp>
#include <field.hpp>
#include <vsrc.hpp>
#include <horn.hpp>
#include <domain.hpp>
#include <cstdint>

#define PYBIND11_DETAILED_ERROR_MESSAGES 1

namespace py = pybind11;

using namespace QOSM::Propagation;
using namespace QOSM::Pose;
using namespace QOSM::Item;
using namespace QOSM::EM;
using namespace QOSM::EM::Source;

VectorNx3 array_to_VectorNx3(const py::array_t<double> &vectors){
    // convert arguments from python to C++
    py::buffer_info info = vectors.request();
    if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
        throw std::runtime_error("Vector list must have shape (N,3)");
    }
    VectorNx3 points(info.shape[0]);
    double *data = points.data();
    const double *in = vectors.data();
    for (int i=0; i<info.shape[0]*3; i++){
        data[i] = in[i];
    }
    return points;
}

py::array_t<cplx> field_to_numpy(const FieldNx3& arr) {
    py::array_t<cplx> res({(size_t)arr.numRows(), (size_t)arr.numCols()});
    const cplx *data = arr.data();
    cplx *ptr = res.mutable_data();
    auto size = arr.size();
    for (size_t i=0; i<size; i++)
        ptr[i] = data[i];

    return res;
}

py::array_t<double> vector_to_numpy(const VectorNx3& arr) {
    py::array_t<double> res({(size_t)arr.numRows(), (size_t)arr.numCols()});
    const double *data = arr.data();
    double *ptr = res.mutable_data();
    auto size = arr.size();
    for (size_t i=0; i<size; i++)
        ptr[i] = data[i];

    return res;
}

py::array_t<double> array_to_numpy(const nc::NdArray<double>& arr) {
    py::array_t<double> res({(size_t)arr.numRows(), (size_t)arr.numCols()});
    const double *data = arr.data();
    double *ptr = res.mutable_data();
    auto size = arr.size();
    for (size_t i=0; i<size; i++)
        ptr[i] = data[i];

    return res;
}

py::array_t<double> array_to_flatten_numpy(const nc::NdArray<double>& arr) {
    py::array_t<double> res({(size_t)(arr.numRows() * arr.numCols())});
    const double *data = arr.data();
    double *ptr = res.mutable_data();
    auto size = arr.size();
    for (size_t i=0; i<size; i++)
        ptr[i] = data[i];

    return res;
}

py::array_t<cplx> array_to_numpy(const nc::NdArray<cplx>& arr) {
    py::array_t<cplx> res({(size_t)arr.numRows(), (size_t)arr.numCols()});
    const cplx *data = arr.data();
    cplx *ptr = res.mutable_data();
    auto size = arr.size();
    for (size_t i=0; i<size; i++)
        ptr[i] = data[i];

    return res;
}

py::array_t<double> array_to_numpy(const Vec3& arr) {
    py::array_t<double> res({(size_t)(3)});
    double *ptr = res.mutable_data();
    for (size_t i=0; i<3; i++)
        ptr[i] = arr[i];

    return res;
}


PYBIND11_MODULE(qosm_core, m) {


    py::class_<Vec3>(m, "Vec3")
        .def(py::init<const double&, const double&, const double&>(), py::arg("x")=0., py::arg("y")=0., py::arg("z")=0.)
        .def(py::init<const std::array<double, 3>&>(), py::arg("array"))
        .def_property_readonly("x", &Vec3::getx)
        .def_property_readonly("y", &Vec3::gety)
        .def_property_readonly("z", &Vec3::getz)
        .def("dot", &Vec3::dot, "Dot product between 2 Vec3")
        .def("cross", &Vec3::cross, "Dot product between 2 Vec3")
        .def("norm", &Vec3::norm, "Norm of the vector")
        .def("norm2", &Vec3::norm2, "Square norm of the vector")
        .def("normalise", &Vec3::normalise, "Normalise the vector")
        .def("normalised", &Vec3::normalised, "Return a normalised vector")
        .def("min", &Vec3::min, "Minimum value of the vector")
        .def("max", &Vec3::max, "Maximum value of the vector")
        .def("minimum", &Vec3::minimum, "Return a vector where each component is the min of 2 vectors")
        .def("maximum", &Vec3::maximum, "Return a vector where each component is the max of 2 vectors")
        .def("argmin", &Vec3::argmin, "Get the index of the minimal value")
        .def("argmax", &Vec3::argmax, "Get the index of the maximal value")
        .def("tuple", [](const Vec3& self) { return py::make_tuple(self[0], self[1], self[2]); }, "Get the vector as a tuple")
        .def("toNd", &Vec3::toNd, "Get the vector as a VectorNx3", py::arg("N"))
        .def("numpy",[](const Vec3 &self) {
            return array_to_numpy(self);
        }, "Return the data as a np.ndarray of shape (N,)")
        /*Addition*/
        .def("__add__", [](const Vec3& lhl, const Vec3& rhl) { return lhl + rhl; }, py::is_operator())
        .def("__add__", [](const Vec3& lhl, const double rhl) { return lhl + rhl; }, py::is_operator())
        .def("__add__", [](const double lhl, const Vec3& rhl) { return lhl + rhl; }, py::is_operator())
        .def("__iadd__", [](Vec3& lhl, const Vec3& rhl) { lhl += rhl; return lhl; }, py::is_operator())
        .def("__iadd__", [](Vec3& lhl, const double& rhl) { lhl += rhl; return lhl; }, py::is_operator())
        /*Substraction*/
        .def("__sub__", [](const Vec3& lhl, const Vec3& rhl) { return lhl - rhl; }, py::is_operator())
        .def("__sub__", [](const Vec3& lhl, const double rhl) { return lhl - rhl; }, py::is_operator())
        .def("__sub__", [](const double lhl, const Vec3& rhl) { return lhl - rhl; }, py::is_operator())
        .def("__isub__", [](Vec3& lhl, const Vec3& rhl) { lhl -= rhl; return lhl; }, py::is_operator())
        .def("__isub__", [](Vec3& lhl, const double& rhl) { lhl -= rhl; return lhl; }, py::is_operator())
        /*Multiplication*/
        .def("__mul__", [](const Vec3& lhl, const Vec3& rhl) { return lhl * rhl; }, py::is_operator())
        .def("__mul__", [](const Vec3& lhl, const double rhl) { return lhl * rhl; }, py::is_operator())
        .def("__mul__", [](const double lhl, const Vec3& rhl) { return lhl * rhl; }, py::is_operator())
        .def("__imul__", [](Vec3& lhl, const Vec3& rhl) { lhl *= rhl; return lhl; }, py::is_operator())
        .def("__imul__", [](Vec3& lhl, const double& rhl) { lhl *= rhl; return lhl; }, py::is_operator())
        /*Division*/
        .def("__div__", [](const Vec3& lhl, const Vec3& rhl) { return lhl / rhl; }, py::is_operator())
        .def("__div__", [](const Vec3& lhl, const double rhl) { return lhl / rhl; }, py::is_operator())
        .def("__div__", [](const double lhl, const Vec3& rhl) { return lhl / rhl; }, py::is_operator())
        .def("__idiv__", [](Vec3& lhl, const Vec3& rhl) { lhl /= rhl; return lhl; }, py::is_operator())
        .def("__idiv__", [](Vec3& lhl, const double& rhl) { lhl /= rhl; return lhl; }, py::is_operator())
        /*Subscripts*/
        .def("__getitem__", [](const Vec3& v, size_t index) {
            if (index >= 3) {
                throw py::index_error();
            }
            return v[index];
        })
        .def("__setitem__", [](Vec3& v, size_t index, double value) {
            if (index >= 3) {
                throw py::index_error();
            }
            v[index] = value;
        })
        .def("__str__", &Vec3::__str__);

    py::class_<Quaternion>(m, "Quaternion")
        .def(py::init<const double&, const double&, const double&, const double&>(), py::arg("qw")=1.0, py::arg("qx")=0., py::arg("qy")=0., py::arg("qz")=0.)
        .def(py::init<const std::array<double, 4>&>(), py::arg("array"))
        .def(py::init<const Vec3&, const double&, const bool&>(), py::arg("axis"), py::arg("angle"), py::arg("deg") = false)
        .def_property_readonly("x", &Quaternion::getx)
        .def_property_readonly("y", &Quaternion::gety)
        .def_property_readonly("z", &Quaternion::getz)
        .def_property_readonly("w", &Quaternion::getw)
        //.def_property_readonly("shape", [](const Vec3 &v){return py::make_tuple((size_t)v.numRows(), (size_t)v.numCols());})
        .def("conj", &Quaternion::conj, "Conjugate")
        .def("dot", &Quaternion::dot, "Dot product between 2 quaternion")
        .def("norm", &Quaternion::norm, "Norm of the quaternion")
        .def("norm2", &Quaternion::norm2, "Square norm of the quaternion")
        .def("normalise", &Quaternion::normalise, "Normalise the quaterni²on")
        .def("normalised", &Quaternion::normalised, "Return a normalised quaternion")
        .def("rotate", [](const Quaternion& self, const Vec3& v) { return self.rotate(v);}, "Return a vector")
        .def("rotate", [](const Quaternion& self, const VectorNx3& v) { return self.rotate(v);}, "Return vectors")
        /*Addition*/
        .def("__add__", [](const Quaternion& lhl, const Quaternion& rhl) { return lhl + rhl; }, py::is_operator())
        /*Substraction*/
        .def("__sub__", [](const Quaternion& lhl, const Quaternion& rhl) { return lhl - rhl; }, py::is_operator())
        /*Multiplication*/
        .def("__mul__", [](const Quaternion& lhl, const Quaternion& rhl) { return lhl * rhl; }, py::is_operator())
        .def("__mul__", [](const Quaternion& lhl, const double rhl) { return lhl * rhl; }, py::is_operator())
        .def("__mul__", [](const double lhl, const Quaternion& rhl) { return lhl * rhl; }, py::is_operator())
        /*Division*/
        .def("__div__", [](const Quaternion& lhl, const double rhl) { return lhl / rhl; }, py::is_operator())
        /*Subscripts*/
        .def("__getitem__", [](const Quaternion& v, size_t index) {
            if (index >= 4) {
                throw py::index_error();
            }
            return v[index];
        })
        .def("__setitem__", [](Quaternion& v, size_t index, double value) {
            if (index >= 4) {
                throw py::index_error();
            }
            v[index] = value;
        })
        .def("__str__", &Quaternion::__str__);
    
    py::class_<Frame>(m, "Frame")
        .def(py::init<>())
        .def(py::init<const Vec3&, const Vec3&, const double&, const bool&>(), py::arg("ori"), py::arg("axis"), py::arg("angle"), py::arg("deg") = false)
        .def(py::init<Vec3, Vec3, Vec3, Vec3>(), py::arg("x"), py::arg("y"), py::arg("z"), py::arg("ori"))
        .def(py::init<const Vec3&, const Quaternion&>(), py::arg("ori"), py::arg("quaternion"))
        .def_readonly("x", &Frame::x)
        .def_readonly("y", &Frame::y)
        .def_readonly("z", &Frame::z)
        .def_readonly("ori", &Frame::ori)
        .def_readonly("R", &Frame::R)        
        .def("offset_origin", [](Frame& self, const double &offset){
            self.ori += self.z * offset;
        })
        .def("to_local", (Vec3 (Frame::*)(const Vec3&) const) &Frame::toLocal, "Write a Vec3 in Local Frame")
        .def("to_ref", (Vec3 (Frame::*)(const Vec3&) const) &Frame::toRef, "Write a Vec3 in Reference Frame")
        .def("rot_to_local", (Vec3 (Frame::*)(const Vec3&) const) &Frame::rot2Local, "Rotate a Vec3 into Local Frame")
        .def("rot_to_ref", (Vec3 (Frame::*)(const Vec3&) const) &Frame::rot2Ref, "Rotate a Vec3 Reference Frame")
        .def("to_local", (VectorNx3 (Frame::*)(const VectorNx3&) const) &Frame::toLocal, "Write a Vec3 in Local Frame")
        .def("to_ref", (VectorNx3 (Frame::*)(const VectorNx3&) const) &Frame::toRef, "Write a Vector in Reference Frame")
        .def("rot_to_local", (VectorNx3 (Frame::*)(const VectorNx3&) const) &Frame::rot2Local, "Rotate a Vector into Local Frame")
        .def("rot_to_ref", (VectorNx3 (Frame::*)(const VectorNx3&) const) &Frame::rot2Ref, "Rotate a Vector Reference Frame")

        .def("to_local", [](const Frame &self, const py::array_t<double> &positions){
            const VectorNx3 _positions = array_to_VectorNx3(positions);
            return array_to_numpy(self.toLocal(_positions));
        }, "Write a list of position vectors in Local Frame")
        .def("to_ref", [](const Frame &self, const py::array_t<double> &positions){
            const VectorNx3 _positions = array_to_VectorNx3(positions);
            return array_to_numpy(self.toRef(_positions));
        }, "Write a list of position vectors in Reference Frame")

        .def("rot_to_local", [](const Frame &self, const py::array_t<double> &vectors){
            const VectorNx3 _positions = array_to_VectorNx3(vectors);
            return array_to_numpy(self.rot2Local(_positions));
        }, "Rotate a list of vectors in Local Frame")
        .def("rot_to_ref", [](const Frame &self, const py::array_t<double> &vectors){
            const VectorNx3 _positions = array_to_VectorNx3(vectors);
            return array_to_numpy(self.rot2Ref(_positions));
        }, "Rotate a list of vectors in Local Frame")

        .def("rot_to_local", (FieldNx3 (Frame::*)(const FieldNx3&) const) &Frame::toLocal, "Rotate a Field into Local Frame")
        .def("rot_to_ref", (FieldNx3 (Frame::*)(const FieldNx3&) const) &Frame::toRef, "Rotate a Field into Reference Frame")
        ;
    

    py::class_<QOSM::EM::Source::Source>(m, "Source")
        .def(py::init<double>(), py::arg("frequency_GHz").none(false))
        .def_readonly("freq", &QOSM::EM::Source::Source::freq)
        .def_readonly("k0", &QOSM::EM::Source::Source::k0)
        .def_readonly("lambda0", &QOSM::EM::Source::Source::lambda0);

    py::class_<VirtualSource, QOSM::EM::Source::Source>(m, "VirtualSource")
        .def(py::init<double>(), py::arg("frequency_GHz").none(false))
        .def("set_Ez_is_0_assumption", &VirtualSource::setEzIs0Assumption)
        .def_readwrite("beams", &VirtualSource::beams)
        .def_readonly("freq", &VirtualSource::freq)
        .def_readonly("k0", &VirtualSource::k0)
        .def_readonly("lambda0", &VirtualSource::lambda0)
        .def("init", &VirtualSource::init, "Init vsrc with a list of beams", py::arg("beams").none(false))
        .def("extends", &VirtualSource::extends, "Extends the virtual source with the beam list from another source", py::arg("vsrc").none(false))
        .def("fields", [](const VirtualSource &self, const VectorNx3 &r_pts_src){
                auto values = self.fields(r_pts_src);
                return py::make_tuple(std::get<0>(values), std::get<1>(values), std::get<2>(values));
            }, "Get fields and Poyinting vectors for one or several points")
        .def("oriented_fields", [](const VirtualSource &self, const VectorNx3 &r_pts_src, const Vec3 &z_probe_src){
                auto values = self.orientedFields(r_pts_src, z_probe_src);
                return py::make_tuple(std::get<0>(values), std::get<1>(values), std::get<2>(values));
            }, "Get fields and Poyinting vectors for one or several points")
        .def("e_field", &VirtualSource::eField, "Get E for one or several points")
        .def("h_field", &VirtualSource::hField, "Get H for one or several points")
        .def("fields",[](const VirtualSource &self, const py::array_t<double> &r_pts_src) {
            // convert arguments from python to C++
            py::buffer_info info = r_pts_src.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }
            VectorNx3 points(info.shape[0]);
            double *data = points.data();
            const double *in = r_pts_src.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            // call the mthod
            auto values = self.fields(points);
            // convert outputs to python 
            return py::make_tuple(std::get<0>(values), std::get<1>(values), std::get<2>(values));
        }, py::arg("r_pts_src").none(false))
        .def("oriented_fields",[](const VirtualSource &self, const py::array_t<double> &r_pts_src, const Vec3 &z_probe_src) {
            // convert arguments from python to C++
            py::buffer_info info = r_pts_src.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }
            VectorNx3 points(info.shape[0]);
            double *data = points.data();
            const double *in = r_pts_src.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            // call the mthod
            auto values = self.orientedFields(points, z_probe_src);
            // convert outputs to python 
            return py::make_tuple(std::get<0>(values), std::get<1>(values), std::get<2>(values));
        }, "Get fields and Poyinting vectors for one or several points", py::arg("r_pts_src").none(false), py::arg("z_probe_src").none(false))
        .def("e_field",[](const VirtualSource &self, const py::array_t<double> &r_pts_src) {
            // convert arguments from python to C++
            py::buffer_info info = r_pts_src.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }
            VectorNx3 points(info.shape[0]);
            double *data = points.data();
            const double *in = r_pts_src.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            auto res = self.eField(points);
            std::cout << "Computed " << res.numRows() << " points" << std::endl;
            // call the mthod
            return res;
        }, "Get H for one or several points", py::arg("r_pts_src").none(false))
        .def("h_field",[](const VirtualSource &self, const py::array_t<double> &r_pts_src) {
            // convert arguments from python to C++
            py::buffer_info info = r_pts_src.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }
            VectorNx3 points(info.shape[0]);
            double *data = points.data();
            const double *in = r_pts_src.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            // call the mthod
            return self.hField(points);
        }, "Get E for one or several points", py::arg("r_pts_src").none(false))
        .def("filter",[](VirtualSource &self, const int &type) {
            if (type==0) return self.filter(BeamType::INITIAL);
            else if (type==1) return self.filter(BeamType::REFLECTED);
            else if (type==2) return self.filter(BeamType::TRANSMITTED);
            return self;
        }, "Get a vsrc with beam of type 'type'", py::arg("type").none(false))
        .def("filter", &VirtualSource::filter, "Get a vsrc with beam of type 'type'", py::arg("type").none(false))
        .def("split", &VirtualSource::split, "Split vsrc into a vsrc of initial beams and one of generated beams");
        
    

    py::class_<RectangularAperture>(m, "RectangularAperture")
        .def(py::init<const double&, const double&, const double&, const size_t&, const size_t&, const Frame&>(), 
            py::arg("freq").none(false), py::arg("Lx").none(false), py::arg("Ly").none(false), py::arg("num_modes").none(false), 
            py::arg("num_points").none(false), py::arg("frame").none(false))
        .def_readwrite("Ek", &RectangularAperture::Ek)
        .def_readwrite("Hk", &RectangularAperture::Hk)
        .def_readwrite("frame", &RectangularAperture::frame)
        .def_property_readonly("xa", [](const RectangularAperture &self) {   
            return array_to_numpy(self.xa);
        })
        .def_property_readonly("ya", [](const RectangularAperture &self) {   
            return array_to_numpy(self.ya);
        })
        .def_property_readonly("Xa", [](const RectangularAperture &self) {   
            return array_to_numpy(self.Xa);
        })
        .def_property_readonly("Ya", [](const RectangularAperture &self) {   
            return array_to_numpy(self.Ya);
        })
        .def("propagative_modes", &RectangularAperture::propagativeModes)
        .def("compute_rx", [](const RectangularAperture &self, const Horn &src, const py::array_t<double> &r_pts_src, const bool &onlyCentralPoint) {   
            // convert arguments from python to C++
            py::buffer_info info = r_pts_src.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }         
            VectorNx3 points(info.shape[0]);
            double *data = points.data();
            const double *in = r_pts_src.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            // call the mthod
            const auto src_ptr = std::make_shared<Horn>(src);
            auto ck = self.computeRx(src_ptr, points, onlyCentralPoint);
            return array_to_numpy(ck);
        }, py::arg("src"), py::arg("r_pts_src"), py::arg("only_central_point")=false)
        .def("compute_rx", [](const RectangularAperture &self, const qsrc::VirtualSource &src, const py::array_t<double> &r_pts_src, const bool &onlyCentralPoint) {   
            // convert arguments from python to C++
            py::buffer_info info = r_pts_src.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }         
            VectorNx3 points(info.shape[0]);
            double *data = points.data();
            const double *in = r_pts_src.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            // call the mthod
            const auto src_ptr = std::make_shared<qsrc::VirtualSource>(src);
            auto ck = self.computeRx(src_ptr, points, onlyCentralPoint);
            return array_to_numpy(ck);
        }, py::arg("src"), py::arg("r_pts_src"), py::arg("only_central_point")=false)
        .def("compute_rx", [](const RectangularAperture &self, const FieldNx3 field_src) {
            auto ck = self.computeRx(field_src);
            return array_to_numpy(ck);
        }, py::arg("field_src"))
        .def("e_field_rx", [](const RectangularAperture &self, const Horn &src, const py::array_t<double> &r_pts_src, const bool &onlyCentralPoint) {   
            // convert arguments from python to C++
            py::buffer_info info = r_pts_src.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }         
            VectorNx3 points(info.shape[0]);
            double *data = points.data();
            const double *in = r_pts_src.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            // call the mthod
            const auto src_ptr = std::make_shared<Horn>(src);
            return self.eFieldRx(src_ptr, points, onlyCentralPoint);
        }, py::arg("src"), py::arg("r_pts_src"), py::arg("only_central_point")=false)
        .def("e_field_rx", [](const RectangularAperture &self, const qsrc::VirtualSource &src, const py::array_t<double> &r_pts_src, const bool &onlyCentralPoint) {   
            // convert arguments from python to C++
            py::buffer_info info = r_pts_src.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }         
            VectorNx3 points(info.shape[0]);
            double *data = points.data();
            const double *in = r_pts_src.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            // call the mthod
            const auto src_ptr = std::make_shared<qsrc::VirtualSource>(src);
            return self.eFieldRx(src_ptr, points, onlyCentralPoint);
        }, py::arg("src"), py::arg("r_pts_src"), py::arg("only_central_point")=false);

    py::class_<Horn, QOSM::EM::Source::Source>(m, "Horn")
        .def(py::init<double>(), py::arg("frequency_GHz").none(false))
        .def_readonly("aperture", &Horn::aperture)
        .def_readonly("freq", &Horn::freq)
        .def_readonly("k0", &Horn::k0)
        .def_readonly("lambda0", &Horn::lambda0)
        .def("init", &Horn::init, "Init horn aperture", py::arg("modes").none(false), py::arg("Lx").none(false), py::arg("Ly").none(false), py::arg("rot_z_deg")=0., py::arg("num_pts_aperture")=51)
        .def("apertureFields", &Horn::apertureFields)
        .def("fields", [](const Horn &self, const VectorNx3 &r_pts_src){
                auto values = self.fields(r_pts_src);
                return py::make_tuple(std::get<0>(values), std::get<1>(values), std::get<2>(values));
            }, "Get fields and Poyinting vectors for one or several points")
        .def("oriented_fields", [](const Horn &self, const VectorNx3 &r_pts_src, const Vec3 &z_probe_src){
                auto values = self.orientedFields(r_pts_src, z_probe_src);
                return py::make_tuple(std::get<0>(values), std::get<1>(values), std::get<2>(values));
            }, "Get fields and Poyinting vectors for one or several points")
        .def("e_field", &Horn::eField, "Get E for one or several points")
        .def("h_field", &Horn::hField, "Get H for one or several points")
        .def("fields",[](const Horn &self, const py::array_t<double> &r_pts_src) {
            // convert arguments from python to C++
            py::buffer_info info = r_pts_src.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }
            VectorNx3 points(info.shape[0]);
            double *data = points.data();
            const double *in = r_pts_src.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            // call the mthod
            auto values = self.fields(points);
            // convert outputs to python 
            return py::make_tuple(std::get<0>(values), std::get<1>(values), std::get<2>(values));
        }, py::arg("r_pts_src").none(false))
        .def("fields", &Horn::fields, py::arg("r_pts_src").none(false))
        .def("oriented_fields",[](const Horn &self, const py::array_t<double> &r_pts_src, const Vec3 &z_probe_src) {
            // convert arguments from python to C++
            py::buffer_info info = r_pts_src.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }
            VectorNx3 points(info.shape[0]);
            double *data = points.data();
            const double *in = r_pts_src.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            // call the mthod
            auto values = self.fields(points);
            // convert outputs to python 
            return py::make_tuple(std::get<0>(values), std::get<1>(values), std::get<2>(values));
        }, "Get fields and Poyinting vectors for one or several points", py::arg("r_pts_src").none(false), py::arg("z_probe_src").none(false))
        .def("e_field",[](const Horn &self, const py::array_t<double> &r_pts_src) {
            // convert arguments from python to C++
            py::buffer_info info = r_pts_src.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }
            VectorNx3 points(info.shape[0]);
            double *data = points.data();
            const double *in = r_pts_src.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            auto res = self.eField(points);
            // call the mthod
            return res;
        }, "Get H for one or several points", py::arg("r_pts_src").none(false))
        .def("h_field",[](const Horn &self, const py::array_t<double> &r_pts_src) {
            // convert arguments from python to C++
            py::buffer_info info = r_pts_src.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }
            VectorNx3 points(info.shape[0]);
            double *data = points.data();
            const double *in = r_pts_src.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            // call the mthod
            return self.hField(points);
        }, "Get E for one or several points", py::arg("r_pts_src").none(false));


        py::class_<FieldNx3>(m, "Field")
        .def(py::init<std::uint64_t, cplx>(), py::arg("N") = 0, py::arg("value") = 0.)
        .def_property_readonly("shape", [](const FieldNx3 &v){return py::make_tuple((size_t)v.numRows(), (size_t)v.numCols());})
        .def("magnitude", [](const FieldNx3 &self, const bool &normalise, const bool &db, const double &db_min, const py::tuple &reshape) {
            auto array = array_to_numpy(self.magnitude(normalise, db, db_min));
            if (reshape[0].cast<int>() > 0)
                array = array.reshape({reshape[0].cast<int>(), reshape[1].cast<int>()});
            return array;
        }, "Compute the magnitude of the field for each point", py::arg("normalise")=false, py::arg("db")=false, py::arg("db_min")=-50., py::arg_v("reshape", py::make_tuple(-1, 1)) )
        .def("intensity", [](const FieldNx3 &self, const bool &normalise, const bool &db, const double &db_min, const py::tuple &reshape) {
            auto array = array_to_numpy(self.intensity(normalise, db, db_min));
            if (reshape[0].cast<int>() > 0)
                array = array.reshape({reshape[0].cast<int>(), reshape[1].cast<int>()});
            return array;
        }, "Compute the intensity of the field for each point", py::arg("normalise")=false, py::arg("db")=false, py::arg("db_min")=-50., py::arg_v("reshape", py::make_tuple(-1, 1)))
        .def("phase", [](const FieldNx3 &self, const short &dim, const bool &deg, const bool &unwrap, const py::tuple &reshape) {
            auto array = array_to_numpy(self.phase(dim, deg, unwrap));
            if (reshape[0].cast<int>() > 0)
                array = array.reshape({reshape[0].cast<int>(), reshape[1].cast<int>()});
            return array;
        }, "Compute the phase of the field for each point", py::arg("dim")=-1, py::arg("deg")=false, py::arg("unwrap")=false, py::arg_v("reshape", py::make_tuple(-1, 1)))
        .def("conj", &FieldNx3::conjugate, "Compute the conjugate of the field for each point")
        .def("cross", &FieldNx3::cross, "Cross product")
        .def("dotc",[](const FieldNx3 &self, const FieldNx3 &field) {
            // call the mthod
            auto dot = self.dot(field);
            // convert outputs to python 
            return array_to_numpy(dot);
        }, "Get the dot product between two fields", py::arg("field").none(false))
        .def("dot",[](const FieldNx3 &self, const VectorNx3 &vector) {
            // call the mthod
            auto dot = self.dot(vector);
            // convert outputs to python 
            return array_to_numpy(dot);
        }, "Get the dot product between a field and a vector", py::arg("vector").none(false))
        .def("point", &FieldNx3::row, "Get the field at a specific point", py::arg("index"))
        .def("dim", &FieldNx3::col, "Get x, y or z (0, 1, or 2) for each point", py::arg("index"))
        .def("x",[](const FieldNx3& field) {return field[field.rSlice(), 0];})
        .def("y",[](const FieldNx3& field) {return field[field.rSlice(), 1];})
        .def("z",[](const FieldNx3& field) {return field[field.rSlice(), 2];})
        .def("numpy",[](const FieldNx3 &self) {
            return field_to_numpy(self);
        }, "Return the data as a np.ndarray: Nx3")
        /*Subscripts*/
        .def("__getitem__", [](const FieldNx3& field, std::pair<int, int> indices) {
            int row = indices.first;
            int col = indices.second;
            return field.at(row, col);
        })
        .def("__getitem__", [](const FieldNx3& field, const std::pair<py::slice, size_t> &indices) {
            size_t start, stop, step, slice_length;
            if (!indices.first.compute(field.numRows(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            return array_to_numpy(field(nc::Slice(start, stop, step), indices.second));
        })
        .def("__getitem__", [](const FieldNx3& field, const std::pair<size_t, py::slice> &indices) {
            size_t start, stop, step, slice_length;
            if (!indices.second.compute(field.numCols(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            return FieldNx3(field.at(indices.first, nc::Slice(start, stop, step)));
        })
        .def("__getitem__", [](const FieldNx3& field, const std::pair<py::slice, py::slice> &indices) {
            size_t start1, stop1, step1, slice_length1, start2, stop2, step2, slice_length2;
            if (!indices.first.compute(field.numRows(), &start1, &stop1, &step1, &slice_length1)) {
                throw std::runtime_error("Unable to compute slice1");
            }
            if (!indices.second.compute(field.numCols(), &start2, &stop2, &step2, &slice_length2)) {
                throw std::runtime_error("Unable to compute slice2");
            }
            return FieldNx3(field.at(nc::Slice(start1, stop1, step1), nc::Slice(start2, stop2, step2)));
        })
        .def("__getitem__", [](const FieldNx3& field, const std::pair<py::array_t<bool>, py::slice> &indices) {
            size_t start, stop, step, slice_length;
            nc::NdArray<bool> rowMask(field.numRows(), 1);
            for(int i=0; i<indices.first.size(); i++){
                rowMask[i] = indices.first.at(i);
            }
            auto rowIndices = nc::arange<int>(0, field.numRows()).getByMask(rowMask);
            if (!indices.second.compute(field.numCols(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice2");
            }            
            return FieldNx3(field.at(rowIndices, nc::Slice(start, stop, step)));
        })
        .def("__getitem__", [](const FieldNx3& field, const std::pair<py::array_t<bool>, size_t> &indices) {
            size_t start, stop, step, slice_length;
            nc::NdArray<bool> rowMask(field.numRows(), 1);
            for(int i=0; i<indices.first.size(); i++){
                rowMask[i] = indices.first.at(i);
            }
            auto rowIndices = nc::arange<int>(0, field.numRows()).getByMask(rowMask);
            return array_to_numpy(field.at(rowIndices, indices.second));
        })
        .def("__setitem__", [](FieldNx3& field, const std::pair<int, int> &indices, cplx value) {
            auto &row = indices.first;
            auto &col = indices.second;
            field(row, col) = value;
        })
        .def("__setitem__", [](FieldNx3& field, const std::pair<py::slice, py::slice> &indices, cplx &value) {
            size_t start1, stop1, step1, slice_length1, start2, stop2, step2, slice_length2;
            if (!indices.first.compute(field.numRows(), &start1, &stop1, &step1, &slice_length1)) {
                throw std::runtime_error("Unable to compute slice1");
            }
            if (!indices.second.compute(field.numCols(), &start2, &stop2, &step2, &slice_length2)) {
                throw std::runtime_error("Unable to compute slice2");
            }
            auto row = nc::Slice(start1, stop1, step1);
            auto col = nc::Slice(start2, stop2, step2);
            field(row, col) = value;
        })
        .def("__setitem__", [](FieldNx3& field, const std::pair<py::slice, size_t> indices, cplx &value) {
            size_t start, stop, step, slice_length;
            if (!indices.first.compute(field.numRows(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            auto row = nc::Slice(start, stop, step);
            auto &col = indices.second;
            field(row, col) = value;
        })
        .def("__setitem__", [](FieldNx3& field, const std::pair<size_t, py::slice> indices, cplx &value) {
            size_t start, stop, step, slice_length;
            if (!indices.second.compute(field.numCols(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            auto &row = indices.first;
            auto col = nc::Slice(start, stop, step);
            field(row, col) = value;
        })
        .def("__setitem__", [](FieldNx3& field, const std::pair<size_t, py::slice> indices, std::vector<cplx> &values) {
            size_t start, stop, step, slice_length;
            if (!indices.second.compute(field.numCols(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            auto &row = indices.first;
            auto col = nc::Slice(start, stop, step);
            field(row, col) = values;
        })
        /*Addition*/
        .def("__add__", [](const FieldNx3& lhl, const FieldNx3& rhl) { return FieldNx3(lhl + rhl); }, py::is_operator())
        .def("__add__", [](const FieldNx3& lhl, const cplx& rhl) { return FieldNx3(lhl + rhl); }, py::is_operator())
        /*Substraction*/
        .def("__sub__", [](const FieldNx3& lhl, const FieldNx3& rhl) { return FieldNx3(lhl - rhl); }, py::is_operator())
        .def("__sub__", [](const FieldNx3& lhl, const cplx& rhl) { return FieldNx3(lhl - rhl); }, py::is_operator())
        /*Multiplication*/
        .def("__mul__", [](const FieldNx3& lhl, const FieldNx3& rhl) { return FieldNx3(lhl * rhl); }, py::is_operator())
        .def("__mul__", [](const FieldNx3& lhl, const cplx rhl) { return FieldNx3(lhl * rhl); }, py::is_operator())
        .def("__mul__", [](const cplx &lhl, const FieldNx3& rhl) { return FieldNx3(lhl * rhl); }, py::is_operator())
        /*Division*/
        .def("__div__", [](const FieldNx3& lhl, const cplx rhl) { return FieldNx3(lhl / rhl); }, py::is_operator())
        ;

        //
        py::class_<FieldNx1>(m, "Field1d")
        .def(py::init<std::uint64_t, double>(), py::arg("N") = 0, py::arg("value") = 0.)
        .def(py::init<FieldNx3, std::uint64_t>(), py::arg("field"), py::arg("N") = 0)
        .def_property_readonly("shape", [](const FieldNx1 &v){return py::make_tuple((size_t)v.numRows(), (size_t)v.numCols());})
        .def("magnitude", [](const FieldNx1 &self, const bool &normalise, const bool &db, const double &db_min, const py::tuple &reshape) {
            auto array = array_to_numpy(self.magnitude(normalise, db, db_min));
            if (reshape[0].cast<int>() > 0)
                array = array.reshape({reshape[0].cast<int>(), reshape[1].cast<int>()});
            return array;
        }, "Compute the magnitude of the field for each point", py::arg("normalise")=false, py::arg("db")=false, py::arg("db_min")=-50., py::arg_v("reshape", py::make_tuple(-1, 1)) )
        .def("intensity", [](const FieldNx1 &self, const bool &normalise, const bool &db, const double &db_min, const py::tuple &reshape) {
            auto array = array_to_numpy(self.intensity(normalise, db, db_min));
            if (reshape[0].cast<int>() > 0)
                array = array.reshape({reshape[0].cast<int>(), reshape[1].cast<int>()});
            return array;
        }, "Compute the intensity of the field for each point", py::arg("normalise")=false, py::arg("db")=false, py::arg("db_min")=-50., py::arg_v("reshape", py::make_tuple(-1, 1)))
        .def("phase", [](const FieldNx1 &self, const bool &deg, const bool &unwrap, const py::tuple &reshape) {
            auto array = array_to_numpy(self.phase(deg, unwrap));
            if (reshape[0].cast<int>() > 0)
                array = array.reshape({reshape[0].cast<int>(), reshape[1].cast<int>()});
            return array;
        }, "Compute the phase of the field for each point", py::arg("deg")=false, py::arg("unwrap")=false, py::arg_v("reshape", py::make_tuple(-1, 1)))
        .def("conj", &FieldNx1::conjugate, "Compute the conjugate of the field for each point")
        .def("point", &FieldNx1::row, "Get the field at a specific point", py::arg("index"))
        .def("numpy",[](const FieldNx1 &self) {
            return field_to_numpy(self);
        }, "Return the data as a np.ndarray: Nx3")
        /*Subscripts*/
        .def("__getitem__", [](const FieldNx1& field, std::pair<int, int> indices) {
            int row = indices.first;
            int col = indices.second;
            return field.at(row, col);
        })
        .def("__getitem__", [](const FieldNx1& field, const std::pair<py::slice, size_t> &indices) {
            size_t start, stop, step, slice_length;
            if (!indices.first.compute(field.numRows(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            return array_to_numpy(field(nc::Slice(start, stop, step), indices.second));
        })
        .def("__getitem__", [](const FieldNx1& field, const std::pair<size_t, py::slice> &indices) {
            size_t start, stop, step, slice_length;
            if (!indices.second.compute(field.numCols(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            return FieldNx1(field.at(indices.first, nc::Slice(start, stop, step)));
        })
        .def("__getitem__", [](const FieldNx1& field, const std::pair<py::slice, py::slice> &indices) {
            size_t start1, stop1, step1, slice_length1, start2, stop2, step2, slice_length2;
            if (!indices.first.compute(field.numRows(), &start1, &stop1, &step1, &slice_length1)) {
                throw std::runtime_error("Unable to compute slice1");
            }
            if (!indices.second.compute(field.numCols(), &start2, &stop2, &step2, &slice_length2)) {
                throw std::runtime_error("Unable to compute slice2");
            }
            return FieldNx1(field.at(nc::Slice(start1, stop1, step1), nc::Slice(start2, stop2, step2)));
        })
        .def("__getitem__", [](const FieldNx1& field, const std::pair<py::array_t<bool>, py::slice> &indices) {
            size_t start, stop, step, slice_length;
            nc::NdArray<bool> rowMask(field.numRows(), 1);
            for(int i=0; i<indices.first.size(); i++){
                rowMask[i] = indices.first.at(i);
            }
            auto rowIndices = nc::arange<int>(0, field.numRows()).getByMask(rowMask);
            if (!indices.second.compute(field.numCols(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice2");
            }            
            return FieldNx1(field.at(rowIndices, nc::Slice(start, stop, step)));
        })
        .def("__getitem__", [](const FieldNx3& field, const std::pair<py::array_t<bool>, size_t> &indices) {
            size_t start, stop, step, slice_length;
            nc::NdArray<bool> rowMask(field.numRows(), 1);
            for(int i=0; i<indices.first.size(); i++){
                rowMask[i] = indices.first.at(i);
            }
            auto rowIndices = nc::arange<int>(0, field.numRows()).getByMask(rowMask);
            return array_to_numpy(field.at(rowIndices, indices.second));
        })
        .def("__setitem__", [](FieldNx1& field, const std::pair<int, int> &indices, cplx value) {
            auto &row = indices.first;
            auto &col = indices.second;
            field(row, col) = value;
        })
        .def("__setitem__", [](FieldNx1& field, const std::pair<py::slice, py::slice> &indices, cplx &value) {
            size_t start1, stop1, step1, slice_length1, start2, stop2, step2, slice_length2;
            if (!indices.first.compute(field.numRows(), &start1, &stop1, &step1, &slice_length1)) {
                throw std::runtime_error("Unable to compute slice1");
            }
            if (!indices.second.compute(field.numCols(), &start2, &stop2, &step2, &slice_length2)) {
                throw std::runtime_error("Unable to compute slice2");
            }
            auto row = nc::Slice(start1, stop1, step1);
            auto col = nc::Slice(start2, stop2, step2);
            field(row, col) = value;
        })
        .def("__setitem__", [](FieldNx1& field, const std::pair<py::slice, size_t> indices, cplx &value) {
            size_t start, stop, step, slice_length;
            if (!indices.first.compute(field.numRows(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            auto row = nc::Slice(start, stop, step);
            auto &col = indices.second;
            field(row, col) = value;
        })
        .def("__setitem__", [](FieldNx1& field, const std::pair<size_t, py::slice> indices, cplx &value) {
            size_t start, stop, step, slice_length;
            if (!indices.second.compute(field.numCols(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            auto &row = indices.first;
            auto col = nc::Slice(start, stop, step);
            field(row, col) = value;
        })
        .def("__setitem__", [](FieldNx1& field, const std::pair<size_t, py::slice> indices, std::vector<cplx> &values) {
            size_t start, stop, step, slice_length;
            if (!indices.second.compute(field.numCols(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            auto &row = indices.first;
            auto col = nc::Slice(start, stop, step);
            field(row, col) = values;
        })
        /*Addition*/
        .def("__add__", [](const FieldNx1& lhl, const FieldNx1& rhl) { return FieldNx1(lhl + rhl); }, py::is_operator())
        .def("__add__", [](const FieldNx1& lhl, const cplx& rhl) { return FieldNx1(lhl + rhl); }, py::is_operator())
        /*Substraction*/
        .def("__sub__", [](const FieldNx1& lhl, const FieldNx1& rhl) { return FieldNx1(lhl - rhl); }, py::is_operator())
        .def("__sub__", [](const FieldNx1& lhl, const cplx& rhl) { return FieldNx1(lhl - rhl); }, py::is_operator())
        /*Multiplication*/
        .def("__mul__", [](const FieldNx1& lhl, const FieldNx1& rhl) { return FieldNx1(lhl * rhl); }, py::is_operator())
        .def("__mul__", [](const FieldNx1& lhl, const cplx rhl) { return FieldNx1(lhl * rhl); }, py::is_operator())
        .def("__mul__", [](const cplx &lhl, const FieldNx1& rhl) { return FieldNx1(lhl * rhl); }, py::is_operator())
        /*Division*/
        .def("__div__", [](const FieldNx1& lhl, const cplx& rhl) { return FieldNx1(lhl / rhl); }, py::is_operator())
        ;

        py::class_<VectorNx3>(m, "Vector")
        .def(py::init<std::uint64_t, double>(), py::arg("N") = 0, py::arg("value") = 0)
        .def(py::init([](const py::array_t<double>& py_vector) {
            py::buffer_info info = py_vector.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }
            auto vector = std::make_unique<VectorNx3>(info.shape[0], 0.);
            double *data = vector->data();
            const double *in = py_vector.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            return vector;
        }), py::arg("data"))  
        .def_property_readonly("shape", [](const VectorNx3 &v){return py::make_tuple((size_t)v.numRows(), (size_t)v.numCols());})
        .def("norm",[](VectorNx3 &self) {
            return array_to_numpy(self.norm());
        }, "Compute the norm of each vector")
        .def("norm2", [](VectorNx3 &self) {
            return array_to_numpy(self.norm2());
        }, "Compute the square norm of each vector")
        .def("normalise", &VectorNx3::normalise, "Normalise each vector")
        .def("normalised", &VectorNx3::normalised, "Return a set of normalised vectors")
        .def("cross", &VectorNx3::cross, "Cross product of each vector", py::arg("vector"))
        .def("cross", [](VectorNx3& self, const Vec3& vector) {return self.cross(vector.toNd(1));}, "Cross product of each vector", py::arg("vector"))
        .def("cross", [](VectorNx3& self, const py::array_t<double>& py_vector) {
            py::buffer_info info = py_vector.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }
            VectorNx3 vector(info.shape[0]);
            double *data = vector.data();
            const double *in = py_vector.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            return self.cross(vector);
        }, "Dot product of each vector", py::arg("vector"))
        .def("dot", [](VectorNx3& self, const VectorNx3& vector) {return array_to_numpy(self.dot(vector));}, "Dot product of each vector", py::arg("vector"))
        .def("dot", [](VectorNx3& self, const Vec3& vector) {return array_to_numpy(self.dot(vector.toNd(1)));}, "Dot product of each vector", py::arg("vector"))
        .def("dot", [](VectorNx3& self, const py::array_t<double>& py_vector) {
            py::buffer_info info = py_vector.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }
            VectorNx3 vector(info.shape[0]);
            double *data = vector.data();
            const double *in = py_vector.data();
            for (int i=0; i<info.shape[0]*3; i++){
                data[i] = in[i];
            }
            return array_to_numpy(self.dot(vector));
        }, "Dot product of each vector", py::arg("vector"))
        .def("flip", &VectorNx3::flip, "flip direction of each vector")
        .def("point", &VectorNx3::row, "Get the field at a specific point", py::arg("index"))
        .def("dim", &VectorNx3::col, "Get x, y or z (0, 1, or 2) for each point", py::arg("index"))
        .def("x",[](const VectorNx3& vector) {return vector[vector.rSlice(), 0];})
        .def("y",[](const VectorNx3& vector) {return vector[vector.rSlice(), 1];})
        .def("z",[](const VectorNx3& vector) {return vector[vector.rSlice(), 2];})
        .def("numpy",[](const VectorNx3 &self) {
            return array_to_numpy(self);
        }, "Return the data as a np.ndarray: Nx3")
        /*Subscripts*/
        .def("__getitem__", [](const VectorNx3& field, std::pair<int, int> indices) {
            int row = indices.first;
            int col = indices.second;
            return field.at(row, col);
        })
        .def("__getitem__", [](const VectorNx3& vector, const std::pair<py::slice, size_t> &indices) {
            size_t start, stop, step, slice_length;
            if (!indices.first.compute(vector.numRows(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            auto slice = nc::Slice(start, stop, step);
            return array_to_numpy(vector.at(slice, indices.second));
        })
        .def("__getitem__", [](const VectorNx3& vector, const std::pair<size_t, py::slice> &indices) {
            size_t start, stop, step, slice_length;
            if (!indices.second.compute(vector.numCols(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            return VectorNx3(vector.at(indices.first, nc::Slice(start, stop, step)));
        })
        .def("__getitem__", [](const VectorNx3& vector, const std::pair<py::slice, py::slice> &indices) {
            size_t start1, stop1, step1, slice_length1, start2, stop2, step2, slice_length2;
            if (!indices.first.compute(vector.numRows(), &start1, &stop1, &step1, &slice_length1)) {
                throw std::runtime_error("Unable to compute slice1");
            }
            if (!indices.second.compute(vector.numCols(), &start2, &stop2, &step2, &slice_length2)) {
                throw std::runtime_error("Unable to compute slice2");
            }
            return VectorNx3(vector.at(nc::Slice(start1, stop1, step1), nc::Slice(start2, stop2, step2)));
        })
        .def("__getitem__", [](const VectorNx3& vector, const std::pair<py::array_t<bool>, py::slice> &indices) {
            size_t start, stop, step, slice_length;
            auto ind = nc::arange<int>(0, indices.first.size());
            nc::NdArray<bool> rowMask(indices.first.size(), 1);
            for(int i=0; i<indices.first.size(); i++){
                rowMask[i] = indices.first.at(i);
            }
            auto rowIndices = ind.getByMask(rowMask);
            if (!indices.second.compute(vector.numCols(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice2");
            }            
            return VectorNx3(vector.at(rowIndices, nc::Slice(start, stop, step)));
        })
        .def("__getitem__", [](const VectorNx3& vector, const std::pair<py::array_t<bool>, size_t> &indices) {
            size_t start, stop, step, slice_length;
            auto ind = nc::arange<int>(0, indices.first.size());
            nc::NdArray<bool> rowMask(indices.first.size(), 1);
            for(int i=0; i<indices.first.size(); i++){
                rowMask[i] = indices.first.at(i);
            }
            auto rowIndices = ind.getByMask(rowMask);      
            return array_to_numpy(vector.at(rowIndices, indices.second));
        })
        .def("__setitem__", [](VectorNx3& vector, const std::pair<int, int> &indices, double &value) {
            auto &row = indices.first;
            auto &col = indices.second;
            vector.put(row, col, value);
        })
        .def("__setitem__", [](VectorNx3& vector, const std::pair<py::slice, py::slice> &indices, double &value) {
            size_t start1, stop1, step1, slice_length1, start2, stop2, step2, slice_length2;
            if (!indices.first.compute(vector.numRows(), &start1, &stop1, &step1, &slice_length1)) {
                throw std::runtime_error("Unable to compute slice1");
            }
            if (!indices.second.compute(vector.numCols(), &start2, &stop2, &step2, &slice_length2)) {
                throw std::runtime_error("Unable to compute slice2");
            }
            auto row = nc::Slice(start1, stop1-1, step1);
            auto col = nc::Slice(start2, stop2-1, step2);
            vector.put(row, col, value);
        })
        .def("__setitem__", [](VectorNx3& vector, const std::pair<py::slice, size_t> indices, double &value) {
            size_t start, stop, step, slice_length;
            if (!indices.first.compute(vector.numRows(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            auto row = nc::Slice(start, stop, step);
            auto &col = indices.second;
            vector.put(row, col, value);
        })
        .def("__setitem__", [](VectorNx3& vector, const std::pair<size_t, py::slice> indices, double &value) {
            size_t start, stop, step, slice_length;
            if (!indices.second.compute(vector.numCols(), &start, &stop, &step, &slice_length)) {
                throw std::runtime_error("Unable to compute slice");
            }
            auto &row = indices.first;
            auto col = nc::Slice(start, stop, step);
            vector.put(row, col, value);
        })
        /*Addition*/
        .def("__add__", [](const VectorNx3& lhl, const VectorNx3& rhl) { return VectorNx3(lhl + rhl); }, py::is_operator())
        .def("__add__", [](const VectorNx3& lhl, const double rhl) { return VectorNx3(lhl + rhl); }, py::is_operator())
        .def("__add__", [](const double lhl, const VectorNx3& rhl) { return VectorNx3(lhl + rhl); }, py::is_operator())
        .def("__iadd__", &VectorNx3::operator+=, py::is_operator())
        /*Substraction*/
        .def("__sub__", [](const VectorNx3& lhl, const VectorNx3& rhl) { return VectorNx3(lhl - rhl); }, py::is_operator())
        .def("__sub__", [](const VectorNx3& lhl, const double rhl) { return VectorNx3(lhl - rhl); }, py::is_operator())
        .def("__sub__", [](const double lhl, const VectorNx3& rhl) { return VectorNx3(lhl - rhl); }, py::is_operator())
        .def("__isub__", &VectorNx3::operator-=, py::is_operator())
        /*Multiplication*/
        .def("__mul__", [](const VectorNx3& lhl, const VectorNx3& rhl) { return VectorNx3(lhl * rhl); }, py::is_operator())
        .def("__mul__", [](const VectorNx3& lhl, const double rhl) { return VectorNx3(lhl * rhl); }, py::is_operator())
        .def("__mul__", [](const double lhl, const VectorNx3& rhl) { return VectorNx3(lhl * rhl); }, py::is_operator())
        .def("__imul__", &VectorNx3::operator*=, py::is_operator())
        /*Division*/
        .def("__div__", [](const VectorNx3& lhl, const VectorNx3& rhl) { return VectorNx3(lhl / rhl); }, py::is_operator())
        .def("__div__", [](const VectorNx3& lhl, const double rhl) { return VectorNx3(lhl / rhl); }, py::is_operator())
        .def("__div__", [](const double lhl, const VectorNx3& rhl) { return VectorNx3(lhl / rhl); }, py::is_operator())
        .def("__idiv__", &VectorNx3::operator/=, py::is_operator());

    
    

    py::class_<Medium>(m, "Medium")
        .def(py::init<bool>(), py::arg("is_pec") = false)
        .def(py::init<std::complex<double>, bool>(), py::arg("ior_"), py::arg("is_metallic") = false, "Define a Medium with a complexe index of refraction")
        .def(py::init<std::pair<double, double>, bool>(), py::arg("ior"), py::arg("is_metallic") = false, "Define a Medium with a complex index of refraction")
        .def("set_with_permittivity", [](Medium& self, const std::complex<double>& epsilon_r_) {
            return self.setWithPermittivity(epsilon_r_);
        }, py::arg("epsilon_r_"), "Use complex permittivity to set ior")
        .def("set_with_permittivity", [](Medium& self, const double& epsilon_r_real, const double& epsilon_r_imag) {
            return self.setWithPermittivity(epsilon_r_real, epsilon_r_imag);
        }, py::arg("epsilon_r_real"), py::arg("epsilon_r_imag"), "Use complex permittivity to set ior")
        .def("set_with_conductivity", &Medium::setWithConductivity, py::arg("epsilon_r"), py::arg("conductivity"), py::arg("freq"), "Use permittivity and conductivity to set ior. Valid for ONE frequency")
        .def_readonly("n", &Medium::n, "Real part of the index of refraction")
        .def_readonly("kappa", &Medium::kappa, "Imaginary part of the index of refraction (extinction coefficient)")
        .def_readonly("Z", &Medium::Z_, "Complex impedence of wave travelling in the medium")
        .def_readwrite("metallic", &Medium::isMetallic, "Metallic material")
        .def_readonly("pec", &Medium::isPEC, "PEC material");


    py::class_<Dioptre>(m, "Dioptre")
        .def(py::init<>(), "Define a Dioptre with PEC surface")
        .def(py::init<const Medium&, const Medium&>(), py::arg("mat1"), py::arg("mat2"), "Define a Dioptre using two Media")
        .def(py::init<std::pair<double, double>, std::pair<double, double>, bool>(), py::arg("ior1"), py::arg("ior2"), py::arg("is_metallic") = false, "Define a Dioptre with the complex indexes of refraction")
        .def_readwrite("n1", &Dioptre::n1, "Real part of ior for external material")
        .def_readwrite("n2", &Dioptre::n2, "Real part of ior for internal material")
        .def_readwrite("kappa1", &Dioptre::kappa1, "Extinction coefficient for external material")
        .def_readwrite("kappa2", &Dioptre::kappa2, "Extinction coefficientr for external material")
        .def_readwrite("metallic", &Dioptre::isMetallic, "Metallic interface (no transmission)")
        .def_readwrite("pec", &Dioptre::isPEC, "PEC interface (no transmission, pure reflection)");

    
    py::class_<Ray>(m, "Ray")
        .def(py::init<>())
        .def_readwrite("dir", &Ray::dir)
        .def_readwrite("ori", &Ray::ori);
   

    py::class_<Beam>(m, "Beam")
        .def(py::init<>())
        .def(py::init<const double&, const array2d&, const array2c&, const array2d&, const array2d&, const double&, const double&, const double&, const array2d&, const array2i&, const Frame&, const int&>(),
            py::arg("freq"), py::arg("ior"), py::arg("polarisation"), py::arg("w0"), py::arg_v("z0", array2d{0, 0}), 
            py::arg_v("a0", 1.), py::arg_v("phi0", 0.), py::arg_v("initial_power", 0.), py::arg_v("z_range", array2d{-1e30, 1e30}), py::arg_v("bounces", array2i{0, 0}),
            py::arg_v("frame", Frame()), py::arg_v("tracking_id", -1)
        )
        .def(py::init<const double&, const array2d&, const array2c&, const array2c&, const cplx&, const double&, const double&, const double&, const array2d&, const array2i&, const Frame&, const int&>(),
            py::arg("freq"), py::arg("ior"), py::arg("polarisation"), py::arg("w0"), py::arg_v("theta", 0.), 
            py::arg_v("a0", 1.), py::arg_v("phi0", 0.), py::arg_v("initial_power", 0.), py::arg_v("z_range", array2d{-1e30, 1e30}), py::arg_v("bounces", array2i{0, 0}),
            py::arg_v("frame", Frame()), py::arg_v("tracking_id", -1)
        )
        .def("u", [](Beam& self, const Vec3& r_pt_beam) {
            py::array_t<cplx> Q_(std::vector<ptrdiff_t>{2, 2});
            cplx *Q_ptr = Q_.mutable_data();
            std::array<cplx, 4> Q_arr_;
            auto res = gaussianScalarFn(Q_arr_, self, r_pt_beam);
            cplx u_ = res[0];
            Q_ptr[0] = Q_arr_[0];
            Q_ptr[1] = Q_arr_[1];
            Q_ptr[2] = Q_arr_[2];
            Q_ptr[3] = Q_arr_[3];
            return std::make_pair(u_, Q_);
        },  py::arg("r_pt_beam").none(false))
        .def("u", [](Beam& self, const py::array_t<double>& r) {
            py::buffer_info info = r.request();
            if (info.shape.size() < 2 || info.shape.size() > 2 || info.shape[1] != 3){
                throw std::runtime_error("Vector must have shape (N,3)");
            }      
            const int N = info.shape[0];
            VectorNx3 r_pts_beam(N);
            double *data = r_pts_beam.data();
            const double *in = r.data();

            // set output
            py::array_t<cplx> u_(N);
            py::array_t<cplx> Q_(std::vector<ptrdiff_t>{2, 2, N});
            cplx *u_ptr = u_.mutable_data();
            cplx *Q_ptr = Q_.mutable_data();

            int n = 0;
            for (int i=0; i<N*3; i+=3){
                Vec3 r_pt_beam(data[i], data[i+1], data[i+2]);
                std::array<cplx, 4> Q_arr_;
                auto res = gaussianScalarFn(Q_arr_, self, r_pt_beam);
                u_ptr[i/3] = res[0];
                Q_ptr[n++] = Q_arr_[0];
                Q_ptr[n++] = Q_arr_[1];
                Q_ptr[n++] = Q_arr_[2];
                Q_ptr[n++] = Q_arr_[3];
            }
            return std::make_pair(u_, Q_);
        },  py::arg("r_pt_beam").none(false))
        .def_readwrite("frame", &Beam::frame)
        .def("offset_origin", [](Beam& self, const double &offset){
            self.frame.ori += self.frame.z * offset;
        })
        .def_readwrite("k0", &Beam::k0)
        .def_readwrite("q1", &Beam::q1_)
        .def_readwrite("q2", &Beam::q2_)
        .def_readwrite("z0", &Beam::z0)
        .def_readwrite("zr", &Beam::zr)
        .def_readwrite("a0", &Beam::a0)
        .def_readwrite("phi0", &Beam::phi0)
        .def_readwrite("init_power", &Beam::initialPower)
        .def_readwrite("n", &Beam::n)
        .def_readwrite("kappa", &Beam::kappa)
        .def_readwrite("theta", &Beam::theta_)
        .def_readwrite("alpha", &Beam::alpha_)
        .def_readwrite("fresnel", &Beam::fresnel_)
        .def_readwrite("z_range", &Beam::zRange)
        .def_readwrite("bounces_left", &Beam::bouncesLeft)
        .def_property_readonly("beam_type", [](const Beam &self){return (int)self.beamType;})
        .def_readwrite("trck_id", &Beam::trck_id)
        .def("compute_fields", &Beam::computeFields, py::arg("r_pt_src").none(false), py::arg("no_z")=false);


    py::class_<Triangle>(m, "Triangle")
        .def(py::init<const std::array<Vec3, 3>&, const std::array<Vec3, 3>&, const std::array<Vec3, 2>&, const std::array<double, 3>&, const Dioptre&, const int&>(), 
                py::arg("vertices").none(false), py::arg("normals").none(false), py::arg("edges").none(false), py::arg("curvatures").none(false), py::arg("dioptre").none(false), 
                py::arg("id_triangle").none(false))
        .def_readwrite("v0", &Triangle::v0)
        .def_readwrite("v1", &Triangle::v1)
        .def_readwrite("v2", &Triangle::v2)
        .def_readwrite("n0", &Triangle::n0)
        .def_readwrite("n1", &Triangle::n1)
        .def_readwrite("n2", &Triangle::n2)
        .def_readwrite("c0", &Triangle::c0)
        .def_readwrite("c1", &Triangle::c1)
        .def_readwrite("c2", &Triangle::c2)
        .def_readwrite("normal", &Triangle::n)
        .def_readwrite("barycentre", &Triangle::b)
        .def_readwrite("curvature", &Triangle::c)
        .def_readwrite("dioptre", &Triangle::dioptre)
        .def_readwrite("bounces", &Triangle::bounces);

    
    py::enum_<QOSM::Tools::PlaneType>(m, "PlaneType", py::arithmetic())
        .value("XY", QOSM::Tools::PlaneType::XY)
        .value("XZ", QOSM::Tools::PlaneType::XZ)
        .value("YX", QOSM::Tools::PlaneType::YZ)
        .value("YZ", QOSM::Tools::PlaneType::YZ)
        .value("ZX", QOSM::Tools::PlaneType::ZX)
        .value("ZY", QOSM::Tools::PlaneType::ZY)
        .export_values();

    py::class_<QOSM::Tools::Grid>(m, "Grid")  
        .def(py::init<const std::array<double, 3>&, const std::array<double, 3>&, const double&, const QOSM::Tools::PlaneType&>(), 
             py::arg("u_range").none(false),  py::arg("v_range").none(false),  py::arg("n").none(false),  py::arg("plane").none(false), "Create a grid with axis ranges and step")
        .def(py::init<const std::array<double, 2>&, const std::array<double, 2>&, const std::array<int, 2>&, const double&, const QOSM::Tools::PlaneType&>(), 
             py::arg("u_range").none(false),  py::arg("v_range").none(false),  py::arg("num_points").none(false),  py::arg("n").none(false),  py::arg("plane").none(false), "Create a grid with axis ranges and number of points")
        .def_readonly("points", &QOSM::Tools::Grid::points)
        .def_property_readonly("u", [](const QOSM::Tools::Grid &self){ return array_to_flatten_numpy(self.u); }, "u-axis")
        .def_property_readonly("v", [](const QOSM::Tools::Grid &self){ return array_to_flatten_numpy(self.v); }, "v-axis")
        .def_property_readonly("shape", [](const QOSM::Tools::Grid &self){return py::make_tuple(static_cast<int>(self.shape[0]), static_cast<int>(self.shape[1]));}
            , "Number of points along each axis")
        .def("rotate", &QOSM::Tools::Grid::rotate, "Rotate the grid using a quaternion", py::arg("centre"), py::arg("quaternion"))
        .def("rotate_u", &QOSM::Tools::Grid::rotateAlongU, "Rotate the grid along its u-axis", py::arg("centre"), py::arg("angle"), py::arg("deg") = false)
        .def("rotate_v", &QOSM::Tools::Grid::rotateAlongV, "Rotate the grid along its u-axis", py::arg("centre"), py::arg("angle"), py::arg("deg") = false);

    py::class_<Domain>(m, "Domain")   
        .def(py::init<const std::vector<qitem::Triangle>, const int>(), py::arg("geometry").none(false), py::arg("dim")=-1)  
        .def("set_dim", &Domain::setDimension, py::arg("dim"), "Set dimension")
        .def("set_gbt_settings", &Domain::setGBTsettings, py::arg("num_bouces"),  py::arg("threshold_power"), "Set GBT settings")
        .def("propagates", [](Domain &self, const qsrc::Horn& src, const QOSM::Tools::Grid &gbeGrid, const double &samplingStep, const double& kappa){
                const auto src_ptr = std::make_shared<qsrc::Horn>(src);
                return self.propagates(src_ptr, gbeGrid, samplingStep, kappa);
            }, 
            "Performs the propagation inside the domain.",
            py::arg("source"), py::arg("gbe_grid"), py::arg("sampling_step"), py::arg("kappa")
        ).def("propagates", [](Domain &self, const qsrc::VirtualSource& src, const QOSM::Tools::Grid &gbeGrid, const double &samplingStep, const double& kappa){
                const auto src_ptr = std::make_shared<qsrc::VirtualSource>(src);
                return self.propagates(src_ptr, gbeGrid, samplingStep, kappa);
            }, 
            "Performs the propagation inside the domain.",
            py::arg("source"), py::arg("gbe_grid"), py::arg("sampling_step"), py::arg("kappa")
        )
        .def("propagates", [](Domain &self, const qsrc::Horn& src, const VectorNx3 &src_r_gbe, const double &samplingStep, const double& kappa){
                const auto src_ptr = std::make_shared<qsrc::Horn>(src);
                return self.propagates(src_ptr, src_r_gbe, samplingStep, kappa);
            }, 
            "Performs the propagation inside the domain.",
            py::arg("source"), py::arg("src_r_gbe"), py::arg("sampling_step"), py::arg("kappa")
        ).def("propagates", [](Domain &self, const qsrc::VirtualSource& src, const VectorNx3 src_r_gbe, const double &samplingStep, const double& kappa){
                const auto src_ptr = std::make_shared<qsrc::VirtualSource>(src);
                return self.propagates(src_ptr, src_r_gbe, samplingStep,kappa);
            }, 
            "Performs the propagation inside the domain.",
            py::arg("source"), py::arg("src_r_gbe"), py::arg("sampling_step"), py::arg("kappa")
        ).def("propagates", [](Domain &self, const qsrc::VirtualSource& vsrc){
                return self.propagates(vsrc);
            }, 
            "Performs the beam tracing inside the domain.",
            py::arg("vsrc")
        );



    m.def("create_beam", [](
        double freq, array2d ior, array2c polarisation, array2d w0, array2d z0, double a0, double phi0, double initPw, array2d zRange, array2i bounces, Frame frame, int tracking_id
    ) {
        return Beam(
            freq, ior, polarisation, w0, z0, a0, phi0, initPw, zRange, bounces, frame, tracking_id
        );
    }, py::arg("freq").none(false), py::arg("ior").none(false), py::arg("polarisation").none(false), py::arg("w0").none(false), 
    py::arg("z0") = array2d{0., 0.}, py::arg("a0") = 0., py::arg("phi0") = 0., py::arg("init_power") = 0., py::arg("z_range") = array2d{-1e30, 1e30}, py::arg("bounces") = array2i{0, 0}, 
    py::arg("frame") = Frame(), py::arg("tracking_id") = -1);


    m.def("create_beam", [](
        double freq, array2d ior, array2c polarisation, array2c q, cplx theta, double a0, double phi0, double initPw, array2d zRange, array2i bounces, Frame frame, int tracking_id
    ) {
        return Beam(
            freq, ior, polarisation, q, theta, a0, phi0, initPw, zRange, bounces, frame, tracking_id
        );
    }, py::arg("freq").none(false), py::arg("ior").none(false), py::arg("polarisation").none(false), py::arg("q").none(false), 
    py::arg("theta") = array2d{0., 0.}, py::arg("a0") = 0., py::arg("phi0") = 0., py::arg("init_power") = 0., py::arg("z_range") = array2d{-1e30, 1e30}, py::arg("bounces") = array2i{0, 0}, 
    py::arg("frame") = Frame(), py::arg("tracking_id") = -1);

    m.def("frame", [](const py::object &frame, const size_t &index){
        py::array_t<double> ori_arr = frame.attr("ori").cast<py::array_t<double>>();
        py::array_t<double> x_arr = frame.attr("x").cast<py::array_t<double>>();
        py::array_t<double> y_arr = frame.attr("y").cast<py::array_t<double>>();
        py::array_t<double> z_arr = frame.attr("z").cast<py::array_t<double>>();
        Vec3 ori(ori_arr.at(index, 0), ori_arr.at(index, 1), ori_arr.at(index, 2));
        Vec3 x(x_arr.at(index, 0), x_arr.at(index, 1), x_arr.at(index, 2));
        Vec3 y(y_arr.at(index, 0), y_arr.at(index, 1), y_arr.at(index, 2));
        Vec3 z(z_arr.at(index, 0), z_arr.at(index, 1), z_arr.at(index, 2));
        return Frame(x, y, z, ori);
    }
    );


    m.def("field_expansion", [](const py::array_t<cplx>& E, const py::array_t<double>& S, py::array_t<double>& points,  const double& freq, const double& w0, const double &pw_threshold) {
        VectorNx3 poyintingVectors(S.shape()[0]);
        FieldNx3 eFields(E.shape()[0]);
        VectorNx3 beamOrigins(points.shape()[0]);
        for (int i=0; i<points.shape()[0]; i++){
            beamOrigins.put(i, 0, points.at(i, 0));
            beamOrigins.put(i, 1, points.at(i, 1));
            beamOrigins.put(i, 2, points.at(i, 2));
        }       
        for (int i=0; i<S.shape()[0]; i++){
            poyintingVectors.put(i, 0, S.at(i, 0));
            poyintingVectors.put(i, 1, S.at(i, 1));
            poyintingVectors.put(i, 2, S.at(i, 2));
        }      
        for (int i=0; i<E.shape()[0]; i++){
            eFields.put(i, 0, E.at(i, 0));
            eFields.put(i, 1, E.at(i, 1));
            eFields.put(i, 2, E.at(i, 2));
        }
        int N = points.shape()[0];
        auto beams = fieldExpansion(eFields, poyintingVectors, beamOrigins, freq, w0, pw_threshold);
        return beams;
    }, py::arg("field").none(false), py::arg("poyinting").none(false), py::arg("points").none(false), py::arg("freq").none(false), py::arg("w0")=0.001, py::arg("pw_threshold")=1e-5);


    m.def("field_expansion", [](const FieldNx1& E, const VectorNx3& S, const VectorNx3& points,  const double& freq, const int& dim, const double& w0, const double &pw_threshold) {
        return fieldExpansion(E, S, points, freq, dim, w0, pw_threshold);
    }, py::arg("field").none(false), py::arg("poyinting").none(false), py::arg("points").none(false), py::arg("freq").none(false), py::arg("dim"), py::arg("w0")=0.001, py::arg("pw_threshold")=1e-5);

    m.def("field_expansion", [](const FieldNx3& E, const VectorNx3& S, const VectorNx3& points,  const double& freq, const double& w0, const double &pw_threshold) {
        return fieldExpansion(E, S, points, freq, w0, pw_threshold);
    }, py::arg("field").none(false), py::arg("poyinting").none(false), py::arg("points").none(false), py::arg("freq").none(false), py::arg("w0")=0.001, py::arg("pw_threshold")=1e-5);

    m.def("field_expansion", [](const FieldNx3& E, const VectorNx3& S, py::array_t<double>& points,  const double& freq, const double& w0, const double &pw_threshold) {
        VectorNx3 beamOrigins(points.shape()[0]);
        // convert py::array_t to VectorNx3   
        for (int i=0; i<points.shape()[0]; i++){
            beamOrigins.put(i, 0, points.at(i, 0));
            beamOrigins.put(i, 1, points.at(i, 1));
            beamOrigins.put(i, 2, points.at(i, 2));
        }  
        int N = points.shape()[0];
        auto beams = fieldExpansion(E, S, beamOrigins, freq, w0, pw_threshold);
        return beams;
    }, py::arg("field").none(false), py::arg("poyinting").none(false), py::arg("points").none(false), py::arg("freq").none(false), py::arg("w0")=0.001, py::arg("pw_threshold")=1e-5);

    m.def("field_expansion", [](const VirtualSource& vsrc, py::array_t<double>& points,  const double& w0, const double &pw_threshold) {
        VectorNx3 beamOrigins(points.shape()[0]);
        // convert py::array_t to VectorNx3   
        for (int i=0; i<points.shape()[0]; i++){
            beamOrigins.put(i, 0, points.at(i, 0));
            beamOrigins.put(i, 1, points.at(i, 1));
            beamOrigins.put(i, 2, points.at(i, 2));
        }  
        int N = points.shape()[0];
        const auto res = vsrc.fields(beamOrigins);
        const FieldNx3 E = std::get<0>(res);
        const VectorNx3 S = std::get<2>(res);
        const double& freq = vsrc.freq;
        auto beams = fieldExpansion(E, S, beamOrigins, freq, w0, pw_threshold);
        return beams;
    }, py::arg("source").none(false), py::arg("points").none(false), py::arg("w0")=0.001, py::arg("pw_threshold")=1e-5);

    m.def("field_expansion", [](const Horn& horn, py::array_t<double>& points,  const double& w0, const double &pw_threshold) {
        VectorNx3 beamOrigins(points.shape()[0]);
        // convert py::array_t to VectorNx3   
        for (int i=0; i<points.shape()[0]; i++){
            beamOrigins.put(i, 0, points.at(i, 0));
            beamOrigins.put(i, 1, points.at(i, 1));
            beamOrigins.put(i, 2, points.at(i, 2));
        }  
        const auto N = points.shape()[0];
        const auto res = horn.fields(beamOrigins);
        const FieldNx3 E = std::get<0>(res);
        const VectorNx3 S = std::get<2>(res);
        const double& freq = horn.freq;
        auto beams = fieldExpansion(E, S, beamOrigins, freq, w0, pw_threshold);
        return beams;
    }, py::arg("source").none(false), py::arg("points").none(false), py::arg("w0")=0.001, py::arg("pw_threshold")=1e-5);
    
    m.def("field_expansion", [](const VirtualSource& vsrc, const VectorNx3& beamOrigins,  const double& w0, const double &pw_threshold) {
        const auto N = beamOrigins.numRows();
        const auto res = vsrc.fields(beamOrigins);
        const FieldNx3 E = std::get<0>(res);
        const VectorNx3 S = std::get<2>(res);
        const double& freq = vsrc.freq;
        auto beams = fieldExpansion(E, S, beamOrigins, freq, w0, pw_threshold);
        return beams;
    }, py::arg("source").none(false), py::arg("points").none(false), py::arg("w0")=0.001, py::arg("pw_threshold")=1e-5);

    m.def("field_expansion", [](const Horn& horn, const VectorNx3& beamOrigins,  const double& w0, const double &pw_threshold) {
        const auto N = beamOrigins.numRows();
        const auto res = horn.fields(beamOrigins);
        const FieldNx3 E = std::get<0>(res);
        const VectorNx3 S = std::get<2>(res);
        const double& freq = horn.freq;
        auto beams = fieldExpansion(E, S, beamOrigins, freq, w0, pw_threshold);
        return beams;
    }, py::arg("source").none(false), py::arg("points").none(false), py::arg("w0")=0.001, py::arg("pw_threshold")=1e-5);

    m.def("beam_tracing", [](std::vector<qitem::Triangle>& geometry, std::vector<qpropa::Beam>& beams, const array2i &numBouces, const double &thresh, const bool &incrementBouncing) {
        return beamTracing(geometry, beams, numBouces, thresh, incrementBouncing);
    }, py::arg("geometry").none(false), py::arg("beams").none(false), py::arg("bounces").none(false), py::arg("pw_min")=0.00001, py::arg("increment_bouncing")=true);

    m.def("surface_beam_tracing", [](py::array_t<double>& v_normals, py::array_t<double>&v_curvatures, qitem::Dioptre& dioptre, std::vector<qpropa::Beam>& beams, const double &thresh) {
        const double *v_curvatures_ptr = v_curvatures.data();
        std::vector<double> curvatures(v_curvatures_ptr, v_curvatures_ptr + v_curvatures.size());
        std::vector<Vec3> normals(v_normals.shape()[0]);
        for (int i=0; i<v_normals.shape()[0]; i++){
            normals[i] = {v_normals.at(i, 0), v_normals.at(i, 1), v_normals.at(i, 2)};
        }        
        return surfaceBeamTracing(normals, curvatures, dioptre, beams, thresh);
    }, py::arg("normals").none(false), py::arg("curvatures").none(false), py::arg("dioptre").none(false), py::arg("beams").none(false), py::arg("pw_min")=0.001);

    m.def("compute_collected_power", [](const qsrc::VirtualSource &src, py::array_t<double>&aperturePoints, const double &dudv, const qpose::Vec3 &z_probe_src) {
        py::buffer_info info = aperturePoints.request();
        VectorNx3 points(info.shape[0]);
        double *data = points.data();
        const double *in = aperturePoints.data();
        for (int i=0; i<info.shape[0]*3; i++){
            data[i] = in[i];
        }       
        return computeCollectedPower(std::make_shared<qsrc::VirtualSource>(src), points, dudv, z_probe_src);
    }, py::arg("source").none(false), py::arg("aperture_points").none(false), py::arg("dudv").none(false), py::arg("z_probe_src").none(false));

    m.def("compute_collected_power", [](const qsrc::Horn &src, py::array_t<double>&aperturePoints, const double &dudv, const qpose::Vec3 &z_probe_src) {
        py::buffer_info info = aperturePoints.request();
        VectorNx3 points(info.shape[0]);
        double *data = points.data();
        const double *in = aperturePoints.data();
        for (int i=0; i<info.shape[0]*3; i++){
            data[i] = in[i];
        }       
        return computeCollectedPower(std::make_shared<qsrc::Horn>(src), points, dudv, z_probe_src);
    }, py::arg("source").none(false), py::arg("aperture_points").none(false), py::arg("dudv").none(false), py::arg("z_probe_src").none(false));


    #ifdef VERSION_INFO
        m.attr("__version__") = VERSION_INFO;
    #else
        m.attr("__version__") = "dev";
    #endif
}